import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
try:
    import Thermobar as pt
except:
    print('Thermobar cannot be imported, please check your numpy version')

from PetThermoTools.GenFuncs import *
from PetThermoTools.Plotting import *
from PetThermoTools.Liq import *
from PetThermoTools.MELTS import *
from PetThermoTools.Path import *
import multiprocessing
from multiprocessing import Queue
from multiprocessing import Process
import time
from scipy import interpolate
from shapely.geometry import MultiPoint, Point, Polygon

import numpy as np
import pandas as pd
import multiprocessing
import time

def path_4_saturation_multi(q, index, *, Model = None, P_bar = None, comp = None, T_maxdrop_C = None, dt_C = None, T_initial_C = None, fO2_buffer = None,
                      fO2_offset = 0.0, H2O_Sat = None, phases = None):
    """
    Run crystallization simulations at multiple pressures to determine pressure of mineral co-saturations using MELTS or MAGEMinCalc in parallel.

    Parameters
    ----------
    q : multiprocessing.Queue
        Queue object for inter-process communication (used to return results).
    index : array-like
        Index values to iterate over P_bar.
    Model : str
        Thermodynamic model to use.
    P_bar : array-like
        Pressures in bar at which to simulate crystallization paths.
    comp : dict
        Bulk composition of the starting material.
    T_maxdrop_C : float
        Maximum temperature drop below the liquidus and/or starting temperature in Celsius.
    dt_C : float
        Step size for temperature change (in °C).
    T_initial_C : float
        Starting temperature (°C) for simulations.
    fO2_buffer : str, optional
        Oxygen fugacity buffer ("FMQ" or "NNO").
    fO2_offset : float, default 0.0
        Offset in log units from the specified fO2 buffer.
    H2O_Sat : bool
        Whether the system is fluid-saturated.
    phases : list of str
        List of phases to monitor for saturation.

    Returns
    -------
    None
        Results are put in the multiprocessing queue `q` as a list: [idx, results].
    """

    results = {}
    idx = []
    
    if "MELTS" in Model:
        from meltsdynamic import MELTSdynamic

        if Model is None or Model == "MELTSv1.0.2":
            melts = MELTSdynamic(1)
        elif Model == "pMELTS":
            melts = MELTSdynamic(2)
        elif Model == "MELTSv1.1.0":
            melts = MELTSdynamic(3)
        elif Model == "MELTSv1.2.0":
            melts = MELTSdynamic(4)
    else:
        from juliacall import Main as jl, convert as jlconvert

        jl.seval("using MAGEMinCalc")

        comp_julia = jl.seval("Dict")(comp)   

    for i in index:
        # try:
        if "MELTS" in Model:
            Results, tr = path_MELTS(P_bar = P_bar[i],
                    Model=Model,
                    comp=comp,
                    T_maxdrop_C=T_maxdrop_C,
                    dt_C=dt_C,
                    T_initial_C=T_initial_C,
                    find_liquidus=True,
                    fO2_buffer=fO2_buffer,
                    fO2_offset=fO2_offset,
                    fluid_sat=H2O_Sat,
                    Suppress=['rutile', 'tridymite'],
                    phases=phases,
                    trail = True,
                    melts = melts
                )
        else:
            if fO2_offset is None:
                fO2_offset = 0.0

            Results_df = jl.MAGEMinCalc.path(
                                    comp=comp_julia, dt_C=dt_C,
                                    P_bar=P_bar[i], T_min_C = T_maxdrop_C,            
                                    Model=Model, fo2_buffer=fO2_buffer, 
                                    fo2_offset=fO2_offset, find_liquidus=True,
                                    frac_xtal = False
                                    )
            
            Results = dict(Results_df)
            
            tr = True

        idx.append(i)


        results[f"Run {i}"] = Results

        if tr is False:
            break
        # except:
        #     idx.append(i)

    q.put([idx, results])
    return

def mineral_cosaturation(Model="MELTSv1.0.2", cores=int(np.floor(multiprocessing.cpu_count())), bulk=None,
                         phases=['quartz1', 'alkali-feldspar1', 'plagioclase1'],
                         P_bar=np.linspace(250, 5000, 32), Fe3Fet_init=None, H2O_init=None,
                         CO2_init=None, H2O_Sat=False, T_initial_C=None, dt_C=2,
                         T_maxdrop_C=50, T_cut_C=20, find_range=True,
                         find_min=True, fO2_buffer=None, fO2_offset=0.0, timeout = 90, multi_processing = True):
    """
    Determines the pressure at which two or more minerals co-saturate (following the method of Gualda et al. 2014).

    Parameters
    ----------
    Model : str
        Thermodynamic model to use.
    cores : int
        Number of parallel processes to use.
    bulk : dict
        Bulk composition of the starting material.
    phases : list of str
        Mineral phases to track for saturation.
    P_bar : array-like
        Pressures to test in bars.
    Fe3Fet_init : float, optional
        Initial Fe³⁺/Fe_total ratio.
    H2O_init : float, optional
        Initial H₂O content in the system (wt%).
    CO2_init : float, optional
        Initial CO₂ content in the system (wt%).
    H2O_Sat : bool
        If True, run at H₂O saturation.
    T_initial_C : float, optional
        Starting temperature in °C.
    dt_C : float
        Temperature step size (°C).
    T_maxdrop_C : float
        Maximum temperature drop below the liquidus to search for satuartion (°C).
    T_cut_C : float
        Maximum acceptable difference in phase saturation temperatures for "co-saturation" (°C).
    find_range : bool
        If True, analyze phase co-saturation ranges.
    find_min : bool
        If True, determine pressure where phase saturation temperature difference is minimized.
    fO2_buffer : str, optional
        Oxygen fugacity buffer (e.g., "FMQ", "NNO").
    fO2_offset : float, default 0.0
        Offset from the fO2 buffer.
    timeout : int
        Timeout (in seconds) for each process.
    multi_processing : bool
        If True, run simulations in parallel using multiprocessing.

    Returns
    -------
    out : dict
        Contains either 'Output' (DataFrame of phase saturation conditions) or both 'Output' and 'CurveMin'.
    Results : dict
        Raw simulation outputs for each pressure step.
    """
    ## make sure everything is a float
    T_initial_C = to_float(T_initial_C)
    T_maxdrop_C = to_float(T_maxdrop_C)
    T_cut_C = to_float(T_cut_C)

    P_bar      = to_float(P_bar)

    Fe3Fet_init= to_float(Fe3Fet_init)
    H2O_init   = to_float(H2O_init)
    CO2_init   = to_float(CO2_init)
    fO2_offset = to_float(fO2_offset)
    
    comp = bulk.copy()

    if fO2_buffer is not None:
        if fO2_buffer != "NNO":
            if fO2_buffer != "FMQ":
                raise Warning("fO2 buffer specified is not an allowed input. This argument can only be 'FMQ' or 'NNO' \n if you want to offset from these buffers use the 'fO2_offset' argument.")

    if "MELTS" not in Model:
        if fO2_buffer == "FMQ":
            fO2_buffer = "qfm"
        if fO2_buffer == "NNO":
            fO2_buffer = "nno"
            
    if H2O_Sat:
        comp['H2O_Liq'] = 20

    comp = comp_fix(Model=Model, comp=comp, Fe3Fet_Liq=Fe3Fet_init,
                    H2O_Liq=H2O_init, CO2_Liq=CO2_init)

    combined_results = {}
    if multi_processing:
        index_in = np.arange(len(P_bar))
        index_out = np.array([], dtype=int)

        while len(index_out) < len(index_in):
            index = np.setdiff1d(index_in, index_out)
            groups = np.array_split(index, cores)
            non_empty_groups = [g for g in groups if g.size > 0]
            groups = non_empty_groups

            processes = []
            for i in range(len(groups)):
                q = Queue()
                p = Process(target = path_4_saturation_multi, args = (q, groups[i]),
                            kwargs = {'Model': Model, 'comp': comp,
                            'T_initial_C': T_initial_C, 'T_maxdrop_C': T_maxdrop_C,
                            'dt_C': dt_C, 'P_bar': P_bar, 'phases': phases,
                            'fO2_buffer': fO2_buffer, 'fO2_offset': fO2_offset})
                
                processes.append((p, q, groups[i]))
                p.start()

            for p, q, group in processes:
                try:
                    res = q.get(timeout=timeout)
                except:
                    res = []
                    idx_chunks = np.array([group[0]], dtype = int)
                p.join(timeout = 2)
                p.terminate()

                if len(res) > 0.0:
                    idx_chunks, results = res
                    combined_results.update(results)

                index_out = np.hstack((index_out, idx_chunks))
    else:
        if "MELTS" in Model:
            from meltsdynamic import MELTSdynamic

            if Model is None or Model == "MELTSv1.0.2":
                melts = MELTSdynamic(1)
            elif Model == "pMELTS":
                melts = MELTSdynamic(2)
            elif Model == "MELTSv1.1.0":
                melts = MELTSdynamic(3)
            elif Model == "MELTSv1.2.0":
                melts = MELTSdynamic(4)
        else:
            from juliacall import Main as jl, convert as jlconvert

            jl.seval("using MAGEMinCalc")

            comp_julia = jl.seval("Dict")(comp) 

        if "MELTS" in Model:
            for i in range(len(P_bar)):
                results = {}
                Results = path_MELTS(P_bar = P_bar[i],
                        Model=Model,
                        comp=comp,
                        T_maxdrop_C=T_maxdrop_C,
                        dt_C=dt_C,
                        T_initial_C=T_initial_C,
                        find_liquidus=True,
                        fO2_buffer=fO2_buffer,
                        fO2_offset=fO2_offset,
                        fluid_sat=H2O_Sat,
                        Suppress=['rutile', 'tridymite'],
                        phases=phases,
                        melts = melts
                    )
                results[f"index = {i}"] = Results
                combined_results.update(results)
        else:
            if fO2_offset is None:
                fO2_offset = 0.0

            for i in range(len(P_bar)):
                results = {}
                Results_df = jl.MAGEMinCalc.path(
                                        comp=comp_julia, dt_C=dt_C,
                                        P_bar=P_bar[i], T_min_C = T_maxdrop_C,            
                                        Model=Model, fo2_buffer=fO2_buffer, 
                                        fo2_offset=fO2_offset, find_liquidus=True,
                                        frac_xtal = False
                                        )
                
                Results = dict(Results_df)
                results[f"index = {i}"] = Results
                combined_results.update(results)

    results = combined_results

    Results = stich(Res=results, multi=True, Model=Model)

    ## determine the offset between the phases
    if len(phases)  == 3:
        arr = np.zeros((len(Results.keys()), 4))
        arr2 = np.zeros((len(Results.keys()), 4))
        columns = ['P_bar'] + phases + [phases[0] + ' - ' + phases[1], phases[0] + ' - ' + phases[2], phases[1] + ' - ' + phases[2], '3 Phase Saturation']
        for i, r in enumerate(Results.keys()):
            for idx, p in enumerate(phases):
                if p in Results[r]['Mass'].keys():
                    arr[i, idx+1] = Results[r]['Conditions'].loc[Results[r]['Mass'][p] > 0.0, 'T_C'].values[0]
                else:
                    arr[i, idx + 1] = np.nan

            arr[i,0] = Results[r]['Conditions']['P_bar'].loc[0]

        arr2[:, 0] = np.abs(arr[:,1] - arr[:,2])
        arr2[:, 1] = np.abs(arr[:,1] - arr[:,3])
        arr2[:, 2] = np.abs(arr[:,2] - arr[:,3])
        arr2[:, 3] = np.max(arr2[:,0:2], axis = 1)

        r_arr = np.hstack((arr,arr2))

        out = pd.DataFrame(data = r_arr, columns = columns)
        out = out.sort_values('P_bar').reset_index(drop=True)

    else:
        arr = np.zeros((len(Results.keys()),4))
        columns = ['P_bar'] + phases + [phases[0]+' - '+phases[1]]
        for i, r in enumerate(Results.keys()):
            for idx, p in enumerate(phases):
                if p in Results[r]['Mass'].keys():
                    arr[i, idx+1] = Results[r]['Conditions'].loc[Results[r]['Mass'][p] > 0.0, 'T_C'].values[0]
                else:
                    arr[i, idx + 1] = np.nan

            arr[i,0] = Results[r]['Conditions']['P_bar'].loc[0]

        arr[:,3] = np.abs(arr[:,1] - arr[:,2])

        out = pd.DataFrame(data = arr, columns = columns)
        out = out.sort_values('P_bar').reset_index(drop=True)

    if find_min:
        res = findmin(out = out, P_bar = P_bar, T_cut_C = T_cut_C)
        out = {'CurveMin': res, 'Output': out}
    else:
        out = {'Output': out}

    return out, Results

def findmin(out = None, P_bar = None, T_cut_C = None):
    """
    Finds the minimum temperature offset between mineral saturation temperatures as a function of pressure.

    Parameters
    ----------
    out : pd.DataFrame
        DataFrame containing phase saturation temperatures.
    P_bar : array-like
        Pressure values (bar) corresponding to the rows in `out`.
    T_cut_C : float
        Maximum acceptable temperature difference (°C) for co-saturation to be considered valid.

    Returns
    -------
    CurveMin : dict
        Dictionary containing:
            - 'P_min': Pressure of minimum saturation temperature difference.
            - 'Res_min': Minimum temperature difference.
            - 'y_new': Interpolated curve of temperature differences.
            - 'P_new': Pressure array for interpolated curve.
            - 'test': 'Pass' or 'Fail' depending on whether the result meets `T_cut_C`.
    """
    Res = out.copy()

    if '3 Phase Saturation' in list(Res.keys()):
        T_cut_old = T_cut_C
        T_cut_C = T_cut_C + np.nanmin(Res[4:])
        Minimum = list(Res.keys())[4:]
        CurveMin = {}
        for m in Minimum:
            if len(Res[m][~np.isnan(Res[m].values)]) > 2:
                y = Res[m][(~np.isnan(Res[m].values)) & (Res[m].values < T_cut_C)].values
                x = P_bar[(~np.isnan(Res[m].values)) & (Res[m].values < T_cut_C)]

                try:
                    y_new = interpolate.UnivariateSpline(x, y, k = 3)

                    P_new = np.linspace(P_bar[P_bar == np.nanmin(P_bar[(~np.isnan(Res[m].values)) & (Res[m].values < T_cut_C)])], 
                                        P_bar[P_bar == np.nanmax(P_bar[(~np.isnan(Res[m].values)) & (Res[m].values < T_cut_C)])], 200)

                    NewMin = np.nanmin(y_new(P_new))
                    P_min = P_new[np.where(y_new(P_new) == NewMin)][0]
                    if NewMin < T_cut_old:
                        Test = 'Pass'
                    else:
                        Test = 'Fail'

                    CurveMin[m] = {'P_min': P_min, 'Res_min': NewMin, 'y_new': y_new(P_new), 'P_new': P_new, 'test': Test}
                except:
                    try:
                        y_new = interpolate.UnivariateSpline(x, y, k = 2)

                        P_new = np.linspace(P_bar[P_bar == np.nanmin(P_bar[(~np.isnan(Res[m].values)) & (Res[m].values < T_cut_C)])], 
                                            P_bar[P_bar == np.nanmax(P_bar[(~np.isnan(Res[m].values)) & (Res[m].values < T_cut_C)])], 200)

                        NewMin = np.nanmin(y_new(P_new))
                        P_min = P_new[np.where(y_new(P_new) == NewMin)][0]
                        if NewMin < T_cut_old:
                            Test = 'Pass'
                        else:
                            Test = 'Fail'

                        CurveMin[m] = {'P_min': P_min, 'Res_min': NewMin, 'y_new': y_new(P_new), 'P_new': P_new, 'test': Test}
                    except:
                        CurveMin[m] = {'P_min': np.nan, 'Res_min': np.nan, 'y_new': np.nan, 'P_new': np.nan, 'test': 'Fail'}
            else:
                y_new = np.nan
                P_new = np.nan
                NewMin = np.nan
                P_min = np.nan
                Test = 'Fail'
                CurveMin[m] = {'P_min': P_min, 'Res_min': NewMin, 'y_new': y_new, 'P_new': P_new, 'test': Test}

    else:
        CurveMin = {}
        m = Res.keys()[3]
        if len(Res[m][~np.isnan(Res[m])]) > 2:
            y = Res[m][(~np.isnan(Res[m].values)) & (Res[m].values < T_cut_C*2)].values
            x = P_bar[(~np.isnan(Res[m].values)) & (Res[m].values < T_cut_C*2)]

            try:
                y_new = interpolate.UnivariateSpline(x, y, k = 3)
            except:
                y_new = interpolate.UnivariateSpline(x, y, k = 2)

            P_new = np.linspace(P_bar[P_bar == np.nanmin(P_bar[(~np.isnan(Res[m])) & (Res[m] < T_cut_C*2)])], 
                                    P_bar[P_bar == np.nanmax(P_bar[(~np.isnan(Res[m])) & (Res[m] < T_cut_C*2)])], 200)
            
            NewMin = np.nanmin(y_new(P_new))
            P_min = P_new[np.where(y_new(P_new) == NewMin)][0]
            if NewMin < T_cut_C:
                Test = 'Pass'
            else:
                Test = 'Fail'
        else:
            y_new = np.nan
            P_new = np.nan
            NewMin = np.nan
            P_min = np.nan
            Test = 'Fail'

        CurveMin[m] = {'P_min': P_min, 'Res_min': NewMin, 'y_new': y_new, 'P_new': P_new, 'test': Test}

    return CurveMin

def find_mineral_cosaturation(cores = None, Model = None, bulk = None, phases = None, P_bar = None, Fe3Fet_Liq = None, 
                              H2O_Liq = None, H2O_Sat = False, T_initial_C = None, dt_C = None, T_maxdrop_C = None, 
                              T_cut_C = None, find_range = None, find_min = None, fO2_buffer = None, fO2_offset = None):
    print("This function has been removed following update to v0.2.40. Please switch to using the mineral_cosaturation() function")

    return "This function has been removed following update to v0.2.40. Please switch to using the mineral_cosaturation() function"

# def find_mineral_cosaturation(cores = None, Model = None, bulk = None, phases = None, P_bar = None, Fe3Fet_Liq = None, 
#                               H2O_Liq = None, H2O_Sat = False, T_initial_C = None, dt_C = None, T_maxdrop_C = None, 
#                               T_cut_C = None, find_range = None, find_min = None, fO2_buffer = None, fO2_offset = None):
#     '''
#     Carry out multiple calculations in parallel. Allows isobaric, polybaric and isochoric crystallisation to be performed as well as isothermal, isenthalpic or isentropic decompression. All temperature inputs/outputs are reported in degrees celcius and pressure is reported in bars.

#     Parameters:
#     ----------
#     cores: int
#         number of processes to run in parallel. Default will be determined using Multiprocessing.cpu_count().

#     Model: string
#         "MELTS" or "Holland". Dictates whether MELTS or MAGEMin calculations are performed. Default "MELTS".
#         Version of melts can be specified "MELTSv1.0.2", "MELTSv1.1.0", "MELTSv1.2.0", or "pMELTS". Default "v.1.0.2".

#     bulk: Dict or pd.DataFrame
#         Initial compositon for calculations. If type == Dict, the same initial composition will be used in all calculations.

#     phases: list
#         length 2 or 3, contains the phases of the co-saturated magma. Default = ['quartz1', 'plagioclase1', 'k-feldspar1'].

#     P_bar: np.ndarray
#         Calculation pressure. Length determines the number of calculations to be performed.

#     Fe3Fet_Liq: float or np.ndarray
#         Initial Fe 3+/total ratio.

#     H2O_Liq: float
#         H2O content of the initial melt phase.

#     T_initial_C: float
#         Starting temperature for the liquidus calculations. Default = 1200

#     T_maxdrop_C: float
#         Max temperature drop of the calculations. Default = 25

#     dt_C: float
#         Temperature change at each model step. Default = 1

#     T_cut_C: float
#         Temperature offset used to indicate whether the model has succeeded or failed in finding a match. Default = 10.

#     find_range: True/False
#         If True a new DataFrame will be included in Results, indicating all cases where the minimum offset is less than T_cut_C.

#     find_min: True/False
#         If True, a spline fit will be applied to the data to find the minimum point.

#     fO2_buffer: string
#         If the oxygen fugacity of the system is to be buffered during crystallisation/decompression, then an offset to a known buffer must be specified. Here the user can define the known buffer as either "FMQ" or "NNO".

#     fO2_offset: float
#         Offset from the buffer spcified in fO2_buffer (log units).

#     Returns:
#     ----------
#     Results: Dict
#         Dictionary containing information regarding the saturation temperature of each phase and the residuals between the different phases
#     '''
#     print('This function will be removed following the update to v0.3.0. Please switch to using the mineral_cosaturation() function')

#     try:
#         from meltsdynamic import MELTSdynamic
#     except:
#         Warning('alphaMELTS for Python files are not on the python path. \n Please add these files to the path running \n import sys \n sys.path.append(r"insert_your_path_to_melts_here") \n You are looking for the location of the meltsdynamic.py file')

#     T_step_C = dt_C
#     dt_C = T_maxdrop_C 

#     comp = bulk.copy()
#     if H2O_Sat is True:
#         comp['H2O_Liq'] = 20

#     # set default values if required
#     if Model is None:
#         Model == "MELTSv1.0.2"

#     if cores is None:
#         cores = multiprocessing.cpu_count()

#     # if comp is entered as a pandas series, it must first be converted to a dict
#     if type(comp) == pd.core.series.Series:
#         comp = comp.to_dict()

#     # ensure the bulk composition has the correct headers etc.
#     comp = comp_fix(Model = Model, comp = comp)

#     # create base array to be filled
#     if P_bar is None:
#         P_bar = np.array([-1])
#     elif type(P_bar) != np.ndarray:
#         P_bar = np.array([P_bar])

#     if Fe3Fet_Liq is None:
#         Fe3Fet_Liq = np.array([-1])
#     elif type(Fe3Fet_Liq) != np.ndarray:
#         Fe3Fet_Liq = np.array([Fe3Fet_Liq])

#     if H2O_Liq is None:
#         H2O_Liq = np.array([-1])
#     elif type(H2O_Liq) != np.ndarray:
#         H2O_Liq = np.array([H2O_Liq])

#     base_array = np.zeros((len(H2O_Liq),len(Fe3Fet_Liq),len(P_bar)))

#     if P_bar[0] == -1:
#         P_bar = 1000
#     if Fe3Fet_Liq[0] == -1:
#         Fe3Fet_Liq = None
#     if H2O_Liq[0] == -1:
#         H2O_Liq = None

#     # set default values for remaining parameters
#     if T_initial_C is None:
#         T_initial_C = 1200

#     if dt_C is None:
#         dt_C = 25

#     if T_step_C is None:
#         T_step_C = 1

#     if T_cut_C is None:
#         T_cut_C = 10

#     if phases is None:
#         phases = ['quartz1', 'alkali-feldspar1', 'plagioclase1']

#     # create main output dictionary
#     Results = {}
#     if len(phases) == 3:
#         List = [phases[0], phases[1], phases[2], 'T_Liq', 'H2O_melt', '3 Phase Saturation', phases[0] + ' - ' + phases[1], phases[0] + ' - ' + phases[2], phases[1] + ' - ' + phases[2]]
#         for l in List:
#             Results[l] = base_array.copy()
#     else:
#         List = [phases[0], phases[1], 'T_Liq', 'H2O_melt', phases[0] + ' - ' + phases[1]]
#         for l in List:
#             Results[l] = base_array.copy()

#     # run calculations if only one initial composition provided
#     if type(comp) == dict:
#         for i in range(np.shape(base_array)[0]):
#             # if H2O specified set H2O on each iteration
#             if H2O_Liq is not None:
#                 comp['H2O_Liq'] = H2O_Liq[i]

#             for j in range(np.shape(base_array)[1]):
#                 #if Fe3Fet specified set Fe3Fet on each iteration
#                 if Fe3Fet_Liq is not None:
#                     comp['Fe3Fet_Liq'] = Fe3Fet_Liq[j]

#                 # determine how many processes to run in parallel
#                 if len(P_bar) > 1:
#                     A = len(P_bar)//cores
#                     B = len(P_bar) % cores

#                 if A > 0:
#                     Group = np.zeros(A) + cores
#                     if B > 0:
#                         Group = np.append(Group, B)
#                 else:
#                     Group = np.array([B])

#                 # initialise queue
#                 qs = []
#                 q = Queue()


#                 # run calculations
#                 for k in range(len(Group)):
#                     ps = []

#                     for kk in range(int(cores*k), int(cores*k + Group[k])):
#                         p = Process(target = satTemperature, args = (q, kk),
#                                     kwargs = {'Model': Model, 'comp': comp,
#                                     'T_initial_C': T_initial_C, 'T_step_C': T_step_C,
#                                     'dt_C': dt_C, 'P_bar': P_bar[kk], 'phases': phases,
#                                     'H2O_Liq': H2O_Liq, 'fO2_buffer': fO2_buffer, 'fO2_offset': fO2_offset})

#                         ps.append(p)
#                         p.start()

#                     TIMEOUT = 300

#                     start = time.time()
#                     for p in ps:
#                         if time.time() - start < TIMEOUT - 10:
#                             try:
#                                 ret = q.get(timeout = TIMEOUT - (time.time()-start) + 10)
#                             except:
#                                 ret = []
#                         else:
#                             try:
#                                 ret = q.get(timeout = 10)
#                             except:
#                                 ret = []

#                         qs.append(ret)

#                     TIMEOUT = 5
#                     start = time.time()
#                     for p in ps:
#                         if p.is_alive():
#                             while time.time() - start <= TIMEOUT:
#                                 if not p.is_alive():
#                                     p.join()
#                                     p.terminate()
#                                     break
#                                 time.sleep(.1)
#                             else:
#                                 p.terminate()
#                                 p.join(5)
#                         else:
#                             p.join()
#                             p.terminate()


#                     # for p in ps:
#                     #     try:
#                     #         ret = q.get(timeout = 180)
#                     #     except:
#                     #         ret = []
#                     #
#                     #     qs.append(ret)
#                     #
#                     # TIMEOUT = 20
#                     # start = time.time()
#                     # for p in ps:
#                     #     if p.is_alive():
#                     #         time.sleep(.1)
#                     #         while time.time() - start <= TIMEOUT:
#                     #             if not p.is_alive():
#                     #                 p.join()
#                     #                 p.terminate()
#                     #                 break
#                     #             time.sleep(.1)
#                     #         else:
#                     #             p.terminate()
#                     #             p.join(5)
#                     #     else:
#                     #         p.join()
#                     #         p.terminate()

#                 # # extract results
#                 for kk in range(len(qs)):
#                     if len(qs[kk]) > 0:
#                         Res, index = qs[kk]
#                         for l in Results:
#                             if l != 'sat_surface':
#                                 Results[l][i,j,index] = Res[l]

#         # covert any empty values to nan
#         for l in Results:
#             if l != 'sat_surface':
#                 Results[l][np.where(Results[l] == 0.0)] = np.nan

#         if find_min is not None:
#             if H2O_Liq is not None:
#                 Results = findMinimum(Results = Results, P_bar = P_bar, T_cut_C = T_cut_C, H2O_Liq = H2O_Liq, Fe3Fet_Liq = Fe3Fet_Liq, phases = phases)
#             else:
#                 Results = findMinimum(Results = Results, P_bar = P_bar, T_cut_C = T_cut_C, H2O_Liq = H2O_Liq, Fe3Fet_Liq = Fe3Fet_Liq, phases = phases)

#         if find_range is not None:
#             Results['range'] = np.zeros(np.shape(Results[l]))
#             if len(phases) == 3:
#                 Results['range'][np.where(Results['3 Phase Saturation'] <= T_cut_C)] = True
#                 Results['range'][np.where(Results['3 Phase Saturation'] > T_cut_C)] = False
#             else:
#                 Results['range'][np.where(Results[phases[0] + ' - ' + phases[1]] <= T_cut_C)] = True
#                 Results['range'][np.where(Results[phases[0] + ' - ' + phases[1]] > T_cut_C)] = False

    

#     return Results

# def findMinimum(Results = None, P_bar = None, T_cut_C = None, H2O_Liq = None, Fe3Fet_Liq = None, phases = None):
#     '''
#     Take the results of SatPress and search for the minimum point using a spline fit.
#     '''
#     if T_cut_C is None:
#         T_cut_C = 10

#     if H2O_Liq is None and Fe3Fet_Liq is None:
#         if '3 Phase Saturation' in list(Results.keys()):
#             Minimum = {'3 Phase Saturation', phases[0] + ' - ' + phases[1], phases[0] + ' - ' + phases[2], phases[1] + ' - ' + phases[2]}
#             CurveMin = {}
#             for m in Minimum:
#                 if len(Results[m][0,0,:][~np.isnan(Results[m][0,0,:])]) > 2:
#                     y = Results[m][0,0,:][(np.where(~np.isnan(Results[m][0,0,:]))) and (np.where(Results[m][0,0,:] < T_cut_C))]
#                     x = P_bar[(np.where(~np.isnan(Results[m][0,0,:]))) and (np.where(Results[m][0,0,:] < T_cut_C))]

#                     try:
#                         y_new = interpolate.UnivariateSpline(x, y, k = 5)
#                     except:
#                         y_new = interpolate.UnivariateSpline(x, y, k = 2)

#                     P_new = np.linspace(P_bar[np.where(P_bar == np.nanmin(P_bar[(np.where(~np.isnan(Results[m][0,0,:]))) and (np.where(Results[m][0,0,:] < T_cut_C))]))], P_bar[np.where(P_bar == np.nanmax(P_bar[(np.where(~np.isnan(Results[m][0,0,:]))) and (np.where(Results[m][0,0,:] < T_cut_C))]))], 200)

#                     NewMin = np.nanmin(y_new(P_new))
#                     P_min = P_new[np.where(y_new(P_new) == NewMin)][0]
#                     if NewMin < T_cut_C:
#                         Test = 'Pass'
#                     else:
#                         Test = 'Fail'

#                     CurveMin[m] = {'P_min': P_min, 'Res_min': NewMin, 'y_new': y_new(P_new), 'P_new': P_new, 'test': Test}
#                 else:
#                     y_new = np.nan
#                     P_new = np.nan
#                     NewMin = np.nan
#                     P_min = np.nan
#                     Test = 'Fail'
#                     CurveMin[m] = {'P_min': P_min, 'Res_min': NewMin, 'y_new': y_new, 'P_new': P_new, 'test': Test}

#             Results['CurveMin'] = CurveMin
#         else:
#             m = phases[0] + ' - ' + phases[1]
#             if len(Results[m][0,0,:][~np.isnan(Results[m][0,0,:])]) > 2:
#                 y = Results[m][0,0,:][np.where(~np.isnan(Results[m][0,0,:]))]
#                 x = P_bar[np.where(~np.isnan(Results[m][0,0,:]))]

#                 try:
#                     y_new = interpolate.UnivariateSpline(x, y, k=5)
#                 except:
#                     y_new = interpolate.UnivariateSpline(x, y, k = 2)

#                 P_new = np.linspace(P_bar[np.where(P_bar == np.nanmin(P_bar[np.where(~np.isnan(Results[m][0,0,:]))]))], P_bar[np.where(P_bar == np.nanmax(P_bar[np.where(~np.isnan(Results[m][0,0,:]))]))], 200)

#                 NewMin = np.nanmin(y_new(P_new))
#                 P_min = P_new[np.where(y_new(P_new) == NewMin)][0]
#                 if NewMin < T_cut_C:
#                     Test = 'Pass'
#                 else:
#                     Test = 'Fail'
#             else:
#                 y_new = np.nan
#                 P_new = np.nan
#                 NewMin = np.nan
#                 P_min = np.nan
#                 Test = 'Fail'

#             Results['CurveMin'] = {phases[0] + ' - ' + phases[1]: {'P_min': P_min, 'Res_min': NewMin, 'y_new': y_new, 'P_new': P_new, 'test': Test}}

#     if H2O_Liq is not None and Fe3Fet_Liq is None:
#         if '3 Phase Saturation' in list(Results.keys()):
#             X, Y = np.meshgrid(P_bar, H2O_Liq)
#             Y = Results['H2O_melt'][:,0,:].copy()

#             Minimum = {'3 Phase Saturation', phases[0] + ' - ' + phases[1], phases[0] + ' - ' + phases[2], phases[1] + ' - ' + phases[2]}
#             CurveMin = {}
#             for m in Minimum:
#                 if len(Results[m][:,0,:][~np.isnan(Results[m][:,0,:])]) > 4:
#                     Res = Results[m][:,0,:].copy()
#                     Res[np.where(Res > T_cut_C*2)] = np.nan
#                     for i in range(len(H2O_Liq)):
#                         Res[i, :][np.where(Results['H2O_melt'][i,0,:] < 0.99*np.nanmax(Results['H2O_melt'][i,0,:]))] = np.nan

#                     A = Res.copy()
#                     Res[np.where(Res > T_cut_C)] = np.nan
#                     X, Y = np.meshgrid(P_bar, H2O_Liq)
#                     Y = Results['H2O_melt'][:,0,:]

#                     try:
#                         z_new = interpolate.SmoothBivariateSpline(X[np.where(~np.isnan(A) & ~np.isnan(Y))], Y[np.where(~np.isnan(A) & ~np.isnan(Y))], A[np.where(~np.isnan(A) & ~np.isnan(Y))], kx = 3, ky = 3)
#                     except:
#                         z_new = interpolate.SmoothBivariateSpline(X[np.where(~np.isnan(A) & ~np.isnan(Y))], Y[np.where(~np.isnan(A) & ~np.isnan(Y))], A[np.where(~np.isnan(A) & ~np.isnan(Y))], kx = 2, ky = 2)

#                     H2O_new = np.linspace(Y[np.where(Res == np.nanmin(Res))] - (H2O_Liq[1]-H2O_Liq[0]),
#                                         Y[np.where(Res == np.nanmin(Res))] + (H2O_Liq[1]-H2O_Liq[0]), 20)
#                     P_new = np.linspace(X[np.where(Res == np.nanmin(Res))] - (P_bar[1]-P_bar[0]),
#                                         X[np.where(Res == np.nanmin(Res))] + (P_bar[1]-P_bar[0]), 20)

#                     X_new, Y_new = np.meshgrid(P_new, H2O_new)
#                     x = X[~np.isnan(A)].flatten()
#                     y = Y[~np.isnan(A)].flatten()

#                     MyPoly = MultiPoint(list(zip(x, y))).convex_hull

#                     points = list(zip(X_new.flatten(), Y_new.flatten()))
#                     Include = np.zeros(len(X_new.flatten()))
#                     for i in range(len(points)):
#                         p = Point(points[i])
#                         Include[i] = p.within(MyPoly)

#                     YayNay = Include.reshape(X_new.shape)
#                     x_new = X_new[np.where(YayNay == True)].flatten()
#                     y_new = Y_new[np.where(YayNay == True)].flatten()
#                     Res_min = np.nanmin(z_new(x_new, y_new, grid = False))
#                     P_min = x_new[np.where(z_new(x_new, y_new, grid = False) == Res_min)]
#                     H2O_min = y_new[np.where(z_new(x_new, y_new, grid = False) == Res_min)]
#                     if Res_min < T_cut_C:
#                         Test = 'Pass'
#                     else:
#                         Test = 'Fail'

#                     CurveMin[m] = {'Res_min': Res_min, 'P_min': P_min[0], 'H2O_min': H2O_min[0], 'z_new': z_new, 'test': Test}
#                 else:
#                     CurveMin[m] = {'Res_min': np.nan, 'P_min': np.nan, 'H2O_min': np.nan, 'z_new': np.nan, 'test': 'Fail'}

#             Results['CurveMin'] = CurveMin

#     if H2O_Liq is not None and Fe3Fet_Liq is not None:
#         if '3 Phase Saturation' in list(Results.keys()):
#             X, Y = np.meshgrid(P_bar, H2O_Liq)
#             Y = Results['H2O_melt'][:,0,:].copy()

#             Minimum = {'3 Phase Saturation', phases[0] + ' - ' + phases[1], phases[0] + ' - ' + phases[2], phases[1] + ' - ' + phases[2]}
#             CurveMin = {}
#             for m in Minimum:
#                 Res_min_save = 50
#                 for w in range(len(Fe3Fet_Liq)):
#                     if len(Results[m][:,w,:][~np.isnan(Results[m][:,w,:])]) > 4:
#                         Res = Results[m][:,w,:].copy()
#                         Res[np.where(Res > T_cut_C*2)] = np.nan
#                         for i in range(len(H2O_Liq)):
#                             Res[i, :][np.where(Results['H2O_melt'][i,w,:] < 0.99*np.nanmax(Results['H2O_melt'][i,w,:]))] = np.nan

#                         A = Res.copy()
#                         X, Y = np.meshgrid(P_bar, H2O_Liq)
#                         Y = Results['H2O_melt'][:,w,:].copy()

#                         try:
#                             z_new = interpolate.SmoothBivariateSpline(X[np.where(~np.isnan(A) & ~np.isnan(Y))], Y[np.where(~np.isnan(A) & ~np.isnan(Y))], A[np.where(~np.isnan(A) & ~np.isnan(Y))], kx = 3, ky = 3)
#                         except:
#                             z_new = interpolate.SmoothBivariateSpline(X[np.where(~np.isnan(A) & ~np.isnan(Y))], Y[np.where(~np.isnan(A) & ~np.isnan(Y))], A[np.where(~np.isnan(A) & ~np.isnan(Y))], kx = 2, ky = 2)

#                         H2O_new = np.linspace(Y[np.where(Res == np.nanmin(Res))] - (H2O_Liq[1]-H2O_Liq[0]),
#                                             Y[np.where(Res == np.nanmin(Res))] + (H2O_Liq[1]-H2O_Liq[0]), 20)
#                         P_new = np.linspace(X[np.where(Res == np.nanmin(Res))] - (P_bar[1]-P_bar[0]),
#                                             X[np.where(Res == np.nanmin(Res))] + (P_bar[1]-P_bar[0]), 20)

#                         X_new, Y_new = np.meshgrid(P_new, H2O_new)
#                         x = X[~np.isnan(A)].flatten()
#                         y = Y[~np.isnan(A)].flatten()

#                         MyPoly = MultiPoint(list(zip(x, y))).convex_hull

#                         points = list(zip(X_new.flatten(), Y_new.flatten()))
#                         Include = np.zeros(len(X_new.flatten()))
#                         for i in range(len(points)):
#                             p = Point(points[i])
#                             Include[i] = p.within(MyPoly)

#                         YayNay = Include.reshape(X_new.shape)
#                         x_new = X_new[np.where(YayNay == True)].flatten()
#                         y_new = Y_new[np.where(YayNay == True)].flatten()
#                         Res_min = np.nanmin(z_new(x_new, y_new, grid = False))
#                         P_min = x_new[np.where(z_new(x_new, y_new, grid = False) == Res_min)]
#                         H2O_min = y_new[np.where(z_new(x_new, y_new, grid = False) == Res_min)]
#                         if Res_min < T_cut_C:
#                             Test = 'Pass'
#                         else:
#                             Test = 'Fail'

#                         if Res_min < Res_min_save:
#                             Res_min_save = Res_min
#                             CurveMin[m] = {'Res_min': Res_min, 'P_min': P_min[0], 'H2O_min': H2O_min[0], 'z_new': z_new, 'Fe3Fet_Liq': Fe3Fet_Liq[w], 'test': Test}

#             Results['CurveMin'] = CurveMin


#     return Results

# def polymin(P_bar = None, Res = None):
#     '''
#     Finds the minimum residual temperature using a 2nd degree polynomial.
#     '''
#     arr = np.sort(Res)
#     Ind = np.where(Res == arr[0])[0][0]

#     if P_bar[Ind] == np.nanmax(P_bar):
#         p = np.array([0,1,0])
#         p_min = np.array([P_bar[Ind]])
#     elif P_bar[Ind] == np.nanmin(P_bar):
#         p = np.array([0,1,0])
#         p_min = np.array([P_bar[Ind]])
#     else:
#         p = np.polyfit(P_bar[np.array([Ind-1, Ind, Ind+1])],Res[np.array([Ind-1, Ind, Ind+1])],2)

#         x = np.linspace(np.nanmin(P_bar),np.nanmax(P_bar),501)
#         y = p[0]*x**2 + p[1]*x + p[2]

#         p_min = x[np.where(y == np.nanmin(y))]

#     return p, p_min

# def satTemperature(q, index, *, Model = None, comp = None, phases = None, T_initial_C = None, T_step_C = None, dt_C = None, P_bar = None, H2O_Liq = None, fO2_buffer = None, fO2_offset = None):
#     '''
#     Crystallisation calculations to be performed in parallel. Calculations may be either isobaric or isochoric.

#     Parameters:
#     ----------
#     q: Multiprocessing Queue instance
#         Queue instance to record the output variables

#     index: int
#         index of the calculation in the master code (e.g., position within a for loop) to aid indexing results after calculations are complete.

#     Model: string
#         "MELTS" or "Holland". Dictates whether MELTS or MAGEMin calculations are performed. Default "MELTS".
#         Version of melts can be specified "MELTSv1.0.2", "MELTSv1.1.0", "MELTSv1.2.0", or "pMELTS". Default "v.1.0.2".

#     comp: Dict
#         Initial compositon for calculations.

#     T_initial_C: float
#         Initial guess for the liquidus temperature.

#     T_step_C: float
#         The temperature drop at each step of the calculation.

#     dt_C: float
#         Total temperature drop allowed duringmodel runs.

#     P_bar: float
#          Specifies the pressure of calculation (bar).

#     Returns:
#     ----------
#     Results: Dict
#         Dict containing a series of floats that represent the saturation temperature and residual temperature for each calculation.

#     index: int
#         index of the calculation

#     '''

#     Results = {}
#     if "MELTS" in Model:

#         try:
#             Results = phaseSat_MELTS(Model = Model, comp = comp, phases = phases, T_initial_C = T_initial_C, T_step_C = T_step_C, dt_C = dt_C, P_bar = P_bar, H2O_Liq = H2O_Liq, fO2_buffer = fO2_buffer, fO2_offset = fO2_offset)
#         except:
#             Results = {phases[0]: np.nan, phases[1]: np.nan, phases[2]: np.nan, 'T_Liq': np.nan, 'H2O_melt': np.nan}
#             if len(phases) == 2:
#                 del Results[phases[2]]

#         if len(phases) == 3:
#             Res = ['3 Phase Saturation', phases[0] + ' - ' + phases[1], phases[0] + ' - ' + phases[2], phases[1] + ' - ' + phases[2]]
#             for R in Res:
#                 Results[R] = np.nan

#             if ~np.isnan(Results[phases[0]]) and ~np.isnan(Results[phases[1]]) and ~np.isnan(Results[phases[2]]):
#                 Results['3 Phase Saturation'] = np.nanmax(np.array([abs(Results[phases[0]] - Results[phases[1]]), abs(Results[phases[0]] - Results[phases[2]]), abs(Results[phases[1]] - Results[phases[2]])]))
#             if ~np.isnan(Results[phases[0]]) and ~np.isnan(Results[phases[1]]):
#                 Results[phases[0] + ' - ' + phases[1]] = abs(Results[phases[0]] - Results[phases[1]])
#             if ~np.isnan(Results[phases[0]]) and ~np.isnan(Results[phases[2]]):
#                 Results[phases[0] + ' - ' + phases[2]] = abs(Results[phases[0]] - Results[phases[2]])
#             if ~np.isnan(Results[phases[1]]) and ~np.isnan(Results[phases[2]]):
#                 Results[phases[1] + ' - ' + phases[2]] = abs(Results[phases[1]] - Results[phases[2]])
#         else:
#             Results[phases[0] + ' - ' + phases[1]] = np.nan
#             if ~np.isnan(Results[phases[0]]) and ~np.isnan(Results[phases[1]]):
#                 Results[phases[0] + ' - ' + phases[1]] = abs(Results[phases[0]] - Results[phases[1]])

#         q.put([Results, index])
#         return

#     if Model == "Holland":
#         import pyMAGEMINcalc as MM
#         Results = {phases[0]: np.nan, phases[1]: np.nan, phases[2]: np.nan, 'T_Liq': np.nan, 'H2O_melt': np.nan}
#         if len(phases) == 2:
#             del Results[phases[2]]

#         #try:
#         Result = MM.path(comp = comp, phases = phases, T_min_C = dt_C, dt_C = T_step_C, P_bar = P_bar, find_liquidus = True)
#         #Result = stich(Result, Model = Model)

#         for i in range(len(phases)):
#             try:
#                 Results[phases[i]] = Result['Conditions']['T_C'][Result[phases[i]+'_prop']['mass'] > 0.0].values[0]
#                 print(Results[phases[i]])
#             except:
#                 Results[phases[i]] = np.nan

#         Results['T_Liq'] = Result['Conditions']['T_C'].values[0]
#         Results['H2O_melt'] = Result['liq']['H2O'].values[0]

#         if len(phases) == 3:
#             Res = ['3 Phase Saturation', phases[0] + ' - ' + phases[1], phases[0] + ' - ' + phases[2], phases[1] + ' - ' + phases[2]]
#             for R in Res:
#                 Results[R] = np.nan

#             if ~np.isnan(Results[phases[0]]) and ~np.isnan(Results[phases[1]]) and ~np.isnan(Results[phases[2]]):
#                 Results['3 Phase Saturation'] = np.nanmax(np.array([abs(Results[phases[0]] - Results[phases[1]]), abs(Results[phases[0]] - Results[phases[2]]), abs(Results[phases[1]] - Results[phases[2]])]))
#             if ~np.isnan(Results[phases[0]]) and ~np.isnan(Results[phases[1]]):
#                 Results[phases[0] + ' - ' + phases[1]] = abs(Results[phases[0]] - Results[phases[1]])
#             if ~np.isnan(Results[phases[0]]) and ~np.isnan(Results[phases[2]]):
#                 Results[phases[0] + ' - ' + phases[2]] = abs(Results[phases[0]] - Results[phases[2]])
#             if ~np.isnan(Results[phases[1]]) and ~np.isnan(Results[phases[2]]):
#                 Results[phases[1] + ' - ' + phases[2]] = abs(Results[phases[1]] - Results[phases[2]])
#         else:
#             Results[phases[0] + ' - ' + phases[1]] = np.nan
#             if ~np.isnan(Results[phases[0]]) and ~np.isnan(Results[phases[1]]):
#                 Results[phases[0] + ' - ' + phases[1]] = abs(Results[phases[0]] - Results[phases[1]])
        
#         q.put([Results, index])
#         #except:
#         #    q.put([Results, index])
#         return





import numpy as np
import pandas as pd
from PetThermoTools.GenFuncs import *
from PetThermoTools.Plotting import *
from PetThermoTools.MELTS import *
import multiprocessing
from multiprocessing import Queue
from multiprocessing import Process
import time
import sys
from tqdm.notebook import tqdm, trange

try:
    shell = get_ipython().__class__.__name__
    # print(shell)
    if shell == 'ZMQInteractiveShell':
        from wurlitzer import sys_pipes as pipes # Jupyter notebook or qtconsole
    elif shell == 'TerminalInteractiveShell':
        pipes = None  # Terminal running IPython
    else:
        pipes = None  # Other type (?)
except:
    pipes = None     # Probably standard Python interpreter (or wurlitzer not available)

def multi_path(cores = None, Model = None, bulk = None, comp = None, Frac_solid = None, Frac_fluid = None, 
               T_C = None, T_path_C = None, T_start_C = None, T_end_C = None, dt_C = None, 
               P_bar = None, P_path_bar = None, P_start_bar = None, P_end_bar = None, dp_bar = None, 
               Fe3Fet_init = None, Fe3Fet_Liq = None, H2O_init = None, H2O_Liq = None, CO2_init = None, CO2_Liq = None, 
               isenthalpic = None, isentropic = None, isochoric = None, find_liquidus = None, 
               fO2_buffer = None, fO2_offset = None, 
               Print_suppress = None, fluid_sat = None, Crystallinity_limit = None, Combined = None,
               label = None, timeout = None, print_label = True, Suppress = ['rutile', 'tridymite'], Suppress_except=False,
               multi_processing = True):
    """
    Perform single or multiple MELTS/MAGEMin crystallization or degassing calculations in series or parallel.

    Supports isothermal, isenthalpic, isentropic, and isochoric processes over fixed or variable pressure and 
    temperature paths. Works with both MELTS and MAGEMin models. Automatically handles composition expansion when
    using arrays for Fe3+/FeT, H2O, or CO2.

    Parameters
    ----------
    cores : int, optional
        Number of CPU cores to use for multiprocessing. Defaults to total available.
    Model : str
        Thermodynamic model. MELTS variants: "MELTSv1.0.2", "MELTSv1.1.0", "MELTSv1.2.0", "pMELTS";
        or MAGEMin: "Green2025", "Weller2024".
    bulk or comp : dict or pd.DataFrame
        Starting composition(s) for the model run(s). If `dict`, single composition for all runs.
    Frac_solid : bool, optional
        If True, remove solids at each step (fractional crystallization).
    Frac_fluid : bool, optional
        If True, remove fluids at each step.
    T_C, P_bar : float or np.ndarray, optional
        Fixed temperature/pressure values for isothermal/isobaric runs.
    T_path_C, P_path_bar : np.ndarray or 2D array, optional
        User-defined temperature/pressure paths for each model run.
    T_start_C, T_end_C, dt_C : float or array, optional
        Starting, ending, and increment temperature values for path calculations.
    P_start_bar, P_end_bar, dp_bar : float or array, optional
        Starting, ending, and increment pressure values for path calculations.
    Fe3Fet_init : float or array, optional
        Initial Fe3+/FeT ratio(s). Required if redox state is not specified in `comp` or `fO2_buffer`.
    H2O_init, CO2_init : float or array, optional
        Initial melt volatile contents. Required if not specified in `comp`.
    isenthalpic, isentropic, isochoric : bool, optional
        If True, applies respective thermodynamic constraint.
    find_liquidus : bool, optional
        If True, searches for the liquidus before starting.
    fO2_buffer : {"FMQ", "NNO"}, optional
        Specifies redox buffer for constraining oxygen fugacity.
    fO2_offset : float or array, optional
        Offset (in log units) from the specified fO2 buffer.
    Print_suppress : bool, optional
        If True, suppresses status messages.
    fluid_sat : bool, optional
        If True, starts the system at fluid saturation.
    Crystallinity_limit : float, optional
        Ends run when crystallinity (excluding fluids) exceeds this threshold.
    Combined : unused
        Placeholder for backwards compatibility.
    label : str or array, optional
        If provided, labels output dictionaries using `label`.
    timeout : float, optional
        Max allowed runtime for each subprocess (in seconds).
    print_label : bool, default=True
        If True, prints label with status updates.
    Suppress : list of str
        List of phases to exclude from results.
    Suppress_except : bool
        If True, `Suppress` acts as a whitelist instead of a blacklist.
    multi_processing : bool
        If False, run sequentially in main process.

    Returns
    -------
    Results : dict
        Dictionary with each run's label as key. Values are sub-dictionaries of phase/property DataFrames.
        Includes `Input` key summarizing model configuration per run.
    """
    ## make sure everything is a float
    T_C        = to_float(T_C)
    T_path_C   = to_float(T_path_C)
    T_start_C  = to_float(T_start_C)
    T_end_C    = to_float(T_end_C)
    dt_C       = to_float(dt_C)

    P_bar      = to_float(P_bar)
    P_path_bar = to_float(P_path_bar)
    P_start_bar= to_float(P_start_bar)
    P_end_bar  = to_float(P_end_bar)
    dp_bar     = to_float(dp_bar)

    Fe3Fet_init= to_float(Fe3Fet_init)
    Fe3Fet_Liq = to_float(Fe3Fet_Liq)
    H2O_init   = to_float(H2O_init)
    H2O_Liq    = to_float(H2O_Liq)
    CO2_init   = to_float(CO2_init)
    CO2_Liq    = to_float(CO2_Liq)
    fO2_offset = to_float(fO2_offset)

    if timeout is None:
        timeout = 180

    timeout_main = timeout

    if Frac_solid is False:
        Frac_solid = None

    if Frac_fluid is False:
        Frac_fluid = None

    if bulk is not None:
        comp = bulk.copy()
        
    # set default values if required
    if Model is None:
        Model == "MELTSv1.0.2"

    if "MELTS" in Model:
        try:
            from meltsdynamic import MELTSdynamic
        except:
            Warning('alphaMELTS for Python files are not on the python path. \n Please add these files to the path running \n import sys \n sys.path.append(r"insert_your_path_to_melts_here") \n You are looking for the location of the meltsdynamic.py file')

    if H2O_Liq is not None:
        print('Warning - the kwarg "H2O_Liq" will be removed from v1.0.0 onwards. Please use "H2O_init" instead.')
        if H2O_init is None:
            H2O_init = H2O_Liq

    if CO2_Liq is not None:
        print('Warning - the kwarg "CO2_Liq" will be removed from v1.0.0 onwards. Please use "CO2_init" instead.')
        if CO2_init is None:
            CO2_init = CO2_Liq

    if Fe3Fet_Liq is not None:
        print('Warning - the kwarg "Fe3Fet_Liq" will be removed from v1.0.0 onwards. Please use "Fe3Fet_init" instead.')
        if Fe3Fet_init is None:
            Fe3Fet_init = Fe3Fet_Liq

    # if comp is entered as a pandas series, it must first be converted to a dict
    if type(comp) == pd.core.series.Series:
        comp = comp.to_dict()

    if fO2_buffer is not None:
        if fO2_buffer != "NNO":
            if fO2_buffer != "FMQ":
                raise Warning("fO2 buffer specified is not an allowed input. This argument can only be 'FMQ' or 'NNO' \n if you want to offset from these buffers use the 'fO2_offset' argument.")

    if "MELTS" not in Model:
        if fO2_buffer == "FMQ":
            fO2_buffer = "qfm"
        if fO2_buffer == "NNO":
            fO2_buffer = "nno"

    # ensure the bulk composition has the correct headers etc.
    comp = comp_fix(Model = Model, comp = comp, Fe3Fet_Liq = Fe3Fet_init, H2O_Liq = H2O_init, CO2_Liq = CO2_init)
    
    if type(comp) == dict:
        if comp['H2O_Liq'] == 0.0 and "MELTS" in Model:
            raise Warning("Adding small amounts of H$_{2}$O may improve the ability of MELTS to accurately reproduce the saturation of oxide minerals. Additionally, sufficient H$_{2}$O is required in the model for MELTS to predict the crystallisation of apatite, rather than whitlockite.")

        if comp['Fe3Fet_Liq'] == 0.0 and "MELTS" in Model and fO2_buffer is None:
            raise Warning("MELTS often fails to produce any results when no ferric Fe is included in the starting composition and an fO2 buffer is not set.")

    if cores is None:
        cores = multiprocessing.cpu_count()

    # specify the number of calculations to be performed in each sequence
    One = 0
    if type(comp) == pd.core.frame.DataFrame: # simplest scenario - one calculation per bulk composition imported
        A = len(comp['SiO2_Liq'])//cores
        B = len(comp['SiO2_Liq'])%cores
    else:
        if P_bar is not None and type(P_bar) == np.ndarray: #one calculation per P loaded
            A = len(P_bar)//cores
            B = len(P_bar)%cores
        elif T_C is not None and type(T_C) == np.ndarray: # one calculation per T loaded.
            A = len(T_C)//cores
            B = len(T_C)%cores
        elif P_start_bar is not None and type(P_start_bar) == np.ndarray: # one calculation per starting P
            A = len(P_start_bar)//cores
            B = len(P_start_bar)%cores
        elif T_start_C is not None and type(T_start_C) == np.ndarray: # one calculation per starting T
            A = len(T_start_C)//cores
            B = len(T_start_C)%cores
        elif P_path_bar is not None and len(np.shape(P_path_bar)) > 1: # one calculation per P path
            A = np.shape(P_path_bar)[0]//cores
            B = np.shape(P_path_bar)[0]%cores
        elif T_path_C is not None and len(np.shape(T_path_C)) > 1: # one calculation per T path
            A = np.shape(T_path_C)[0]//cores
            B = np.shape(T_path_C)[0]%cores
        elif fO2_offset is not None and type(fO2_offset) == np.ndarray: # one calculation per offset
            A = len(fO2_offset)//cores
            B = len(fO2_offset)%cores

        else: # just one calculation
            One = 1
            A = 1
            B = 0

    Group = np.zeros(A) + cores
    if B > 0:
        Group = np.append(Group, B)

    qs = []
    q = Queue()
    
    # perform calculation if only 1 calculation is specified
    if One == 1:
        if Print_suppress is None:
            print("Running " + Model + " calculation...", end = "", flush = True)
            s = time.time()

        if multi_processing:
            p = Process(target = path, args = (q, 1),
                        kwargs = {'Model': Model, 'comp': comp, 'Frac_solid': Frac_solid, 'Frac_fluid': Frac_fluid,
                                'T_C': T_C, 'T_path_C': T_path_C, 'T_start_C': T_start_C, 'T_end_C': T_end_C, 'dt_C': dt_C,
                                'P_bar': P_bar, 'P_path_bar': P_path_bar, 'P_start_bar': P_start_bar, 'P_end_bar': P_end_bar, 'dp_bar': dp_bar,
                                'isenthalpic': isenthalpic, 'isentropic': isentropic, 'isochoric': isochoric, 'find_liquidus': find_liquidus,
                                'fO2_buffer': fO2_buffer, 'fO2_offset': fO2_offset, 'fluid_sat': fluid_sat, 'Crystallinity_limit': Crystallinity_limit,
                                'Suppress': Suppress, 'Suppress_except': Suppress_except})

            p.start()
            try:
                ret = q.get(timeout = 180)
            except:
                ret = []

            TIMEOUT = 5
            start = time.time()
            if p.is_alive():
                while time.time() - start <= TIMEOUT:
                    if not p.is_alive():
                        p.join()
                        p.terminate()
                        break
                    time.sleep(.1)
                else:
                    p.terminate()
                    p.join(5)
            else:
                p.join()
                p.terminate()

            if Print_suppress is None:
                print(" Complete (time taken = " + str(round(time.time() - s,2)) + " seconds)", end = "", flush = True)

            if len(ret) > 0:
                Results, index = ret
                Results = stich(Results, Model = Model, Frac_fluid = Frac_fluid, Frac_solid = Frac_solid)
                return Results
            else:
                Results = {}
                return Results
        else:
            Results = path_MELTS(Model = Model, comp = comp, Frac_solid = Frac_solid, Frac_fluid = Frac_fluid, 
                                     T_C = T_C, T_path_C = T_path_C, T_start_C = T_start_C, T_end_C = T_end_C, 
                                     dt_C = dt_C, P_bar = P_bar, P_path_bar = P_path_bar, P_start_bar = P_start_bar, 
                                     P_end_bar = P_end_bar, dp_bar = dp_bar, isenthalpic = isenthalpic, 
                                     isentropic = isentropic, isochoric = isochoric, find_liquidus = find_liquidus, 
                                     fO2_buffer = fO2_buffer, fO2_offset = fO2_offset, fluid_sat = fluid_sat, 
                                     Crystallinity_limit = Crystallinity_limit, Suppress = Suppress, Suppress_except = Suppress_except)
            
            Results = stich(Results, Model = Model, Frac_fluid = Frac_fluid, Frac_solid = Frac_solid)
            return Results

    else: # perform multiple crystallisation calculations
        # first make sure everything is the right length
        L = np.sum(Group)
        if P_bar is None:
            P_bar = [None] * int(L)
        elif type(P_bar) == float or type(P_bar) == int:
            P_bar = np.zeros(int(L)) + P_bar
        if T_C is None:
            T_C = [None] * int(L)
        elif type(T_C) == float or type(T_C) == int:
            T_C = np.zeros(int(L)) + T_C

        if P_start_bar is None:
            P_start_bar = [None] * int(L)
        elif type(P_start_bar) == float or type(P_start_bar) == int:
            P_start_bar = np.zeros(int(L)) + P_start_bar
        if T_start_C is None:
            T_start_C = [None] * int(L)
        elif type(T_start_C) == float or type(T_start_C) == int:
            T_start_C = np.zeros(int(L)) + T_start_C

        if P_end_bar is None:
            P_end_bar = [None] * int(L)
        elif type(P_end_bar) == float or type(P_end_bar) == int:
            P_end_bar = np.zeros(int(L)) + P_end_bar
        if T_end_C is None:
            T_end_C = [None] * int(L)
        elif type(T_end_C) == float or type(T_end_C) == int:
            T_end_C = np.zeros(int(L)) + T_end_C

        if dp_bar is None:
            dp_bar = [None] * int(L)
        elif type(dp_bar) == float or type(dp_bar) == int:
            dp_bar = np.zeros(int(L)) + dp_bar
        if dt_C is None:
            dt_C = [None] * int(L)
        elif type(dt_C) == float or type(dt_C) == int:
            dt_C = np.zeros(int(L)) + dt_C

        if P_path_bar is None:
            P_path_bar = [None] * int(L)
        elif len(np.shape(P_path_bar))  == 1:
            P_path_bar = np.vstack([P_path_bar] * int(L))
        if T_path_C is None:
            T_path_C = [None] * int(L)
        elif len(np.shape(T_path_C))  == 1:
            T_path_C = np.vstack([T_path_C] * int(L))

        if fO2_offset is None:
            fO2_offset = [None] * int(L)
        elif type(fO2_offset) == float or type(fO2_offset) == int:
            fO2_offset = np.zeros(int(L)) + fO2_offset

        index_in = np.arange(int(L))
        combined_results = {}
        index_out = np.array([], dtype=int)

        while len(index_out) < len(index_in):
            index = np.setdiff1d(index_in, index_out)
            groups = np.array_split(index, cores)
            non_empty_groups = [g for g in groups if g.size > 0]
            groups = non_empty_groups

            if len(groups[0])< 15:
                timeout = len(groups[0])*timeout_main
            else:
                timeout = 15*timeout_main
                
            processes = []
            Start = time.time()
            for i in range(len(groups)):
                q = Queue()
                p = Process(target = path_multi, args = (q, groups[i]),
                                kwargs = {'Model': Model, 'comp': comp, 'Frac_solid': Frac_solid, 'Frac_fluid': Frac_fluid,
                                        'T_C': T_C, 'T_path_C': T_path_C, 'T_start_C': T_start_C, 'T_end_C': T_end_C, 'dt_C': dt_C,
                                        'P_bar': P_bar, 'P_path_bar': P_path_bar, 'P_start_bar': P_start_bar, 'P_end_bar': P_end_bar, 'dp_bar': dp_bar,
                                        'isenthalpic': isenthalpic, 'isentropic': isentropic, 'isochoric': isochoric, 'find_liquidus': find_liquidus,
                                        'fO2_buffer': fO2_buffer, 'fO2_offset': fO2_offset, 'fluid_sat': fluid_sat, 'Crystallinity_limit': Crystallinity_limit,
                                        'Suppress': Suppress, 'Suppress_except': Suppress_except})
                
                processes.append((p, q, groups[i]))
                p.start()

            for p, q, group in processes:
                try:
                    if time.time() - Start < timeout + 5:
                        res = q.get(timeout=timeout)
                    else:
                        res = q.get(timeout=10)
                except:
                    res = []
                    idx_chunks = np.array([group[0]], dtype = int)

                p.join(timeout = 2)
                p.terminate()

                if len(res) > 0.0:
                    idx_chunks, results = res
                    combined_results.update(results)

                index_out = np.hstack((index_out, idx_chunks))

            print(f"Completed {100*len(index_out)/len(index_in)} %")

        results = combined_results

        Results = stich(Res=results, multi=True, Model=Model, Frac_fluid = Frac_fluid, Frac_solid = Frac_solid)
        
        for r in Results:
            i = int(r.split(' ')[1].strip())
            if type(comp) == dict:
                Results[r]['Input'] = {'Model': Model, 'comp': comp, 'Frac_solid': Frac_solid, 'Frac_fluid': Frac_fluid,
                                    'T_C': T_C[i], 'T_path_C': T_path_C[i], 'T_start_C': T_start_C[i], 'T_end_C': T_end_C[i], 'dt_C': dt_C[i],
                                    'P_bar': P_bar[i], 'P_path_bar': P_path_bar[i], 'P_start_bar': P_start_bar[i], 'P_end_bar': P_end_bar[i], 'dp_bar': dp_bar[i],
                                    'isenthalpic': isenthalpic, 'isentropic': isentropic, 'isochoric': isochoric, 'find_liquidus': find_liquidus,
                                    'fO2_buffer': fO2_buffer, 'fO2_offset': fO2_offset[i], 'fluid_sat': fluid_sat, 'Crystallinity_limit': Crystallinity_limit,
                                    'Suppress': Suppress, 'Suppress_except': Suppress_except}
            else:
                Results[r]['Input'] = {'Model': Model, 'comp': comp.loc[i].to_dict(), 'Frac_solid': Frac_solid, 'Frac_fluid': Frac_fluid,
                                    'T_C': T_C[i], 'T_path_C': T_path_C[i], 'T_start_C': T_start_C[i], 'T_end_C': T_end_C[i], 'dt_C': dt_C[i],
                                    'P_bar': P_bar[i], 'P_path_bar': P_path_bar[i], 'P_start_bar': P_start_bar[i], 'P_end_bar': P_end_bar[i], 'dp_bar': dp_bar[i],
                                    'isenthalpic': isenthalpic, 'isentropic': isentropic, 'isochoric': isochoric, 'find_liquidus': find_liquidus,
                                    'fO2_buffer': fO2_buffer, 'fO2_offset': fO2_offset[i], 'fluid_sat': fluid_sat, 'Crystallinity_limit': Crystallinity_limit,
                                    'Suppress': Suppress, 'Suppress_except': Suppress_except}
        
        if label is not None:
            new_out = label_results(Results,label)
            return new_out
        else:
            return Results

def path_multi(q, index, *, Model = None, comp = None, Frac_solid = None, Frac_fluid = None,
            T_C = None, T_path_C = None, T_start_C = None, T_end_C = None, dt_C = None,
            P_bar = None, P_path_bar = None, P_start_bar = None, P_end_bar = None, dp_bar = None,
            isenthalpic = None, isentropic = None, isochoric = None, find_liquidus = None,
            fO2_buffer = None, fO2_offset = None, fluid_sat = None, Crystallinity_limit = None,
            Suppress = None, Suppress_except = None, trail = True):
    """
    Worker function to run a subset of crystallization/decompression models (MELTS or MAGEMin) in parallel.

    This function is intended to be run in a separate process. It takes a set of indices representing model runs,
    executes them using the appropriate model interface, and returns the results via a multiprocessing queue.

    Parameters
    ----------
    q : multiprocessing.Queue
        Output queue for sending back results.
    index : list of int
        Indices of the simulations to be run by this worker.
    Model : str
        The thermodynamic model ("MELTSv1.0.2", "MELTSv1.1.0", "MELTSv1.2.0", "pMELTS", or MAGEMin variant: "Green2025" or "Weller2024").
    comp : dict or pd.DataFrame
        Starting compositions. Either a single dictionary or a DataFrame with one row per simulation.
    Frac_solid : bool
        If True, removes solids at each step.
    Frac_fluid : bool
        If True, removes fluids at each step.
    T_C, T_path_C, T_start_C, T_end_C, dt_C : float or np.ndarray
        Temperature constraints or paths for each simulation.
    P_bar, P_path_bar, P_start_bar, P_end_bar, dp_bar : float or np.ndarray
        Pressure constraints or paths for each simulation.
    isenthalpic, isentropic, isochoric : bool
        Apply respective thermodynamic constraints.
    find_liquidus : bool
        If True, finds the liquidus temperature before starting.
    fO2_buffer : str
        Oxygen fugacity buffer ("FMQ" or "NNO").
    fO2_offset : float or array
        Offset from specified fO2 buffer in log units.
    fluid_sat : bool
        If True, terminates runs at fluid saturation.
    Crystallinity_limit : float
        Ends run when crystallinity exceeds this value.
    Suppress : list of str
        Phases to exclude from results.
    Suppress_except : bool
        If True, treat `Suppress` as a whitelist.
    trail : bool
        If True, include trailing properties from model output.

    Returns
    -------
    None
        The function returns results using `q.put()`:
        q.put([idx, results])
        where `idx` is the list of completed indices and `results` is a dictionary of output per run.
    """

    results = {}
    idx = []

    if len(index) > 15:
        index = index[:15]
    
    if "MELTS" in Model:
        from meltsdynamic import MELTSdynamic

        if Model is None or Model == "MELTSv1.0.2":
            melts = MELTSdynamic(1)
        elif Model == "pMELTS":
            melts = MELTSdynamic(2)
        elif Model == "MELTSv1.1.0":
            melts = MELTSdynamic(3)
        elif Model == "MELTSv1.2.0":
            melts = MELTSdynamic(4)
    else:
        from juliacall import Main as jl, convert as jlconvert

        jl.seval("using MAGEMinCalc")
  
    for i in index:
        try:
            if "MELTS" in Model:
                if type(comp) == dict:
                    if trail is not None:
                        Results, tr = path_MELTS(Model = Model, comp = comp, Frac_solid = Frac_solid, Frac_fluid = Frac_fluid, 
                                                T_initial_C = 1400, T_C = T_C[i], T_path_C = T_path_C[i], T_start_C = T_start_C[i], 
                                                T_end_C = T_end_C[i], dt_C = dt_C[i], P_bar = P_bar[i], P_path_bar = P_path_bar[i], 
                                                P_start_bar = P_start_bar[i], P_end_bar = P_end_bar[i], dp_bar = dp_bar[i], 
                                                isenthalpic = isenthalpic, isentropic = isentropic, isochoric = isochoric, 
                                                find_liquidus = find_liquidus, fO2_buffer = fO2_buffer, fO2_offset = fO2_offset[i], 
                                                fluid_sat = fluid_sat, Crystallinity_limit = Crystallinity_limit, Suppress = Suppress, 
                                                Suppress_except=Suppress_except, trail = trail, melts = melts)
                    else:
                        Results = path_MELTS(Model = Model, comp = comp, Frac_solid = Frac_solid, Frac_fluid = Frac_fluid, 
                                                T_initial_C = 1400, T_C = T_C[i], T_path_C = T_path_C[i], T_start_C = T_start_C[i], 
                                                T_end_C = T_end_C[i], dt_C = dt_C[i], P_bar = P_bar[i], P_path_bar = P_path_bar[i], 
                                                P_start_bar = P_start_bar[i], P_end_bar = P_end_bar[i], dp_bar = dp_bar[i], 
                                                isenthalpic = isenthalpic, isentropic = isentropic, isochoric = isochoric, 
                                                find_liquidus = find_liquidus, fO2_buffer = fO2_buffer, fO2_offset = fO2_offset[i], 
                                                fluid_sat = fluid_sat, Crystallinity_limit = Crystallinity_limit, Suppress = Suppress, 
                                                Suppress_except=Suppress_except, trail = trail, melts = melts)

                else:
                    if trail is not None:
                        Results, tr = path_MELTS(Model = Model, comp = comp.loc[i].to_dict(), Frac_solid = Frac_solid, Frac_fluid = Frac_fluid, 
                                                T_initial_C = 1400, T_C = T_C[i], T_path_C = T_path_C[i], T_start_C = T_start_C[i], 
                                                T_end_C = T_end_C[i], dt_C = dt_C[i], P_bar = P_bar[i], P_path_bar = P_path_bar[i], 
                                                P_start_bar = P_start_bar[i], P_end_bar = P_end_bar[i], dp_bar = dp_bar[i], 
                                                isenthalpic = isenthalpic, isentropic = isentropic, isochoric = isochoric, 
                                                find_liquidus = find_liquidus, fO2_buffer = fO2_buffer, fO2_offset = fO2_offset[i], 
                                                fluid_sat = fluid_sat, Crystallinity_limit = Crystallinity_limit, Suppress = Suppress, 
                                                Suppress_except=Suppress_except, trail = trail, melts = melts)
                    else:
                        Results = path_MELTS(Model = Model, comp = comp.loc[i].to_dict(), Frac_solid = Frac_solid, Frac_fluid = Frac_fluid, 
                                                T_initial_C = 1400, T_C = T_C[i], T_path_C = T_path_C[i], T_start_C = T_start_C[i], 
                                                T_end_C = T_end_C[i], dt_C = dt_C[i], P_bar = P_bar[i], P_path_bar = P_path_bar[i], 
                                                P_start_bar = P_start_bar[i], P_end_bar = P_end_bar[i], dp_bar = dp_bar[i], 
                                                isenthalpic = isenthalpic, isentropic = isentropic, isochoric = isochoric, 
                                                find_liquidus = find_liquidus, fO2_buffer = fO2_buffer, fO2_offset = fO2_offset[i], 
                                                fluid_sat = fluid_sat, Crystallinity_limit = Crystallinity_limit, Suppress = Suppress, 
                                                Suppress_except=Suppress_except, trail = trail, melts = melts)
            else:
                if fO2_offset[i] is None:
                    fO2_offset[i] = 0.0

                if Frac_solid is None:
                    Frac_solid = False

                if type(comp) == dict:
                    comp_julia = jl.seval("Dict")(comp) 
                else:
                    comp_new = comp.loc[i].to_dict()
                    comp_julia = jl.seval("Dict")(comp_new)
            
                # Results_df = jl.MAGEMinCalc.path(Model = Model, comp=comp_julia, T_start_C = T_start_C[i], T_end_C = T_end_C[i], dt_C=dt_C,
                #                                 T_C = T_C[i], T_path_C = T_path_C[i], P_start_bar = P_start_bar[i], P_end_bar = P_end_bar[i],
                #                                 dp_bar = dp_bar[i], P_bar = P_bar[i], P_path_bar = P_path_bar[i], frac_xtal = Frac_solid,
                #                                 fo2_buffer = fO2_buffer, fo2_offset = fO2_offset[i], find_liquidus = find_liquidus)
                    
                Results_df = jl.MAGEMinCalc.path(
                                    comp=comp_julia, T_start_C=T_start_C[i], T_end_C=T_end_C[i], dt_C=dt_C[i],
                                    T_C=T_C[i], P_start_bar=P_start_bar[i], P_end_bar=P_end_bar[i], dp_bar=dp_bar[i],
                                    P_bar=P_bar[i], T_path_C=T_path_C[i], P_path_bar=P_path_bar[i], frac_xtal=Frac_solid,
                                    Model=Model, fo2_buffer=fO2_buffer, fo2_offset=fO2_offset[i], find_liquidus=find_liquidus
                                )
                
                Results = dict(Results_df)
                
                tr = True

            idx.append(i)


            results[f"Run {i}"] = Results

            if tr is False:
                break
        except:
            idx.append(i)
            break

    q.put([idx, results])
    return


def path(q, index, *, Model = None, comp = None, Frac_solid = None, Frac_fluid = None, 
         T_C = None, T_path_C = None, T_start_C = None, T_end_C = None, dt_C = None, 
         P_bar = None, P_path_bar = None, P_start_bar = None, P_end_bar = None, dp_bar = None, 
         isenthalpic = None, isentropic = None, isochoric = None, find_liquidus = None, 
         fO2_buffer = None, fO2_offset = None, fluid_sat = None, Crystallinity_limit = None,
         Suppress = None, Suppress_except = False):
    '''
    Crystallisation calculations to be performed in parallel. Calculations may be either isobaric or isochoric.

    Parameters:
    ----------
    q: Multiprocessing Queue instance
        Queue instance to record the output variables

    index: int
        index of the calculation in the master code (e.g., position within a for loop) to aid indexing results after calculations are complete.

    Model: string
        "MELTS" or "Holland". Dictates whether MELTS or MAGEMin calculations are performed. Default "MELTS".
        Version of melts can be specified "MELTSv1.0.2", "MELTSv1.1.0", "MELTSv1.2.0", or "pMELTS". Default "v.1.0.2".

    comp: Dict
        Initial compositon for calculations.

    Frac_solid: True/False
        If True, solid phases will be removed from the system at the end of each crystallisation step. Default False.

    Frac_fluid: True/False
        If True, fluid phases will be removed from the system at the end of each crystallisation step. Default False.

    T_C: float
        Calculation temperature - typically used when calculations are performed at a fixed T (e.g.,isothermal degassing).

    T_path_C: np.ndarray
        If a specified temperature path is to be used, T_path_C will be used to determine the T at each step of the model. If 2D, this indicates that multiple calculations with different T_path_C arrays are to be performed.

    T_start_C: float
        Initial temperature used for path calculations.

    T_end_C: float
        Final temperature in crystallisation calculations.

    dt_C: float
        Temperature increment during crystallisation calculations.

    P_bar: float
        Calculation pressure - typically used when calculations are performed at a fixed P (e.g.,isobaric crystallisation).

    P_path_bar: np.ndarray
        If a specified pressure path is to be used, P_path_bar will be used to determine the P at each step of the model. If 2D, this indicates that multiple calculations with different P_path_bar arrays are to be performed.

    P_start_bar: float
        Initial pressure used for path calculations.

    P_end_bar: float
        Final pressure in crystallisation calculations.

    dp_bar: float
        Pressure increment during crystallisation calculations..

    Fe3Fet_Liq: float or np.ndarray
        Fe 3+/total ratio.

    H2O_Liq: float or np.ndarray
        H2O content of the initial bulk composition.

    isenthalpic: True/False
        If True, calculations will be performed at a constant enthalpy with T treated as a dependent variable.

    isentropic: True/False
        If True, calculations will be performed at a constant entropy with T treated as a dependent variable.

    isochoric: True/False
        If True, the volume of the system will be held constant instead of the pressure. Default is False.

    find_liquidus: True/False
        If True, the calculations will start with a search for the melt liquidus temperature. Default is False.

    fO2_buffer: string
        If the oxygen fugacity of the system is to be buffered during crystallisation/decompression, then an offset to a known buffer must be specified. Here the user can define the known buffer as either "FMQ" or "NNO".

    fO2_offset: float
        Offset from the buffer spcified in fO2_buffer (log units).

    Returns:
    ----------
    Results: Dict
        Dict containing a series of pandas DataFrames that display the composition and thermodynamic properties of each phase.

    index: int
        index of the calculation

    '''
    Results = {}
    if "MELTS" in Model:
        try:
            # Results = path_MELTS(Model = Model, comp = comp, Frac_solid = Frac_solid, Frac_fluid = Frac_fluid, T_C = T_C, T_path_C = T_path_C, T_start_C = T_start_C, T_end_C = T_end_C, dt_C = dt_C, P_bar = P_bar, P_path_bar = P_path_bar, P_start_bar = P_start_bar, P_end_bar = P_end_bar, dp_bar = dp_bar, isenthalpic = isenthalpic, isentropic = isentropic, isochoric = isochoric, find_liquidus = find_liquidus, fO2_buffer = fO2_buffer, fO2_offset = fO2_offset, fluid_sat = fluid_sat, Crystallinity_limit = Crystallinity_limit)
            
            if pipes is not None:
                with pipes():
                    Results = path_MELTS(Model = Model, comp = comp, Frac_solid = Frac_solid, Frac_fluid = Frac_fluid, 
                                         T_C = T_C, T_path_C = T_path_C, T_start_C = T_start_C, T_end_C = T_end_C, 
                                         dt_C = dt_C, P_bar = P_bar, P_path_bar = P_path_bar, P_start_bar = P_start_bar, 
                                         P_end_bar = P_end_bar, dp_bar = dp_bar, isenthalpic = isenthalpic, 
                                         isentropic = isentropic, isochoric = isochoric, find_liquidus = find_liquidus, 
                                         fO2_buffer = fO2_buffer, fO2_offset = fO2_offset, fluid_sat = fluid_sat, 
                                         Crystallinity_limit = Crystallinity_limit, Suppress = Suppress, Suppress_except = Suppress_except)
            else:
                Results = path_MELTS(Model = Model, comp = comp, Frac_solid = Frac_solid, Frac_fluid = Frac_fluid, 
                                     T_C = T_C, T_path_C = T_path_C, T_start_C = T_start_C, T_end_C = T_end_C, 
                                     dt_C = dt_C, P_bar = P_bar, P_path_bar = P_path_bar, P_start_bar = P_start_bar, 
                                     P_end_bar = P_end_bar, dp_bar = dp_bar, isenthalpic = isenthalpic, 
                                     isentropic = isentropic, isochoric = isochoric, find_liquidus = find_liquidus, 
                                     fO2_buffer = fO2_buffer, fO2_offset = fO2_offset, fluid_sat = fluid_sat, 
                                     Crystallinity_limit = Crystallinity_limit, Suppress = Suppress, Suppress_except = Suppress_except)
            q.put([Results, index])
        except:
            q.put([])

        return

    if "MELTS" not in Model:
        # import julia
        # from julia.api import Julia
        # jl = Julia(compiled_modules=False)
        # from julia import MAGEMinCalc
        # # import pyMAGEMINcalc as MM
        # # try:
        #     # Results = MM.path(Model = Model, comp = comp, Frac_solid = Frac_solid, Frac_fluid = Frac_fluid, T_C = T_C, T_path_C = T_path_C, T_start_C = T_start_C, T_end_C = T_end_C, dt_C = dt_C, P_bar = P_bar, P_path_bar = P_path_bar, P_start_bar = P_start_bar, P_end_bar = P_end_bar, dp_bar = dp_bar, find_liquidus = find_liquidus, fO2_buffer = fO2_buffer, fO2_offset = fO2_offset)
        # if Frac_solid is None:
        #     Frac_solid = False
            
        # Results = MAGEMinCalc.path(comp = comp, T_start_C = T_start_C, T_end_C = T_end_C, dt_C = dt_C,
        #                             T_C = T_C, P_start_bar = P_start_bar, P_end_bar = P_end_bar, dp_bar = dp_bar,
        #                             P_bar = P_bar, T_path_C = T_path_C, P_path_bar = P_path_bar, frac_xtal = Frac_solid,
        #                             Model = Model, fo2_buffer = fO2_buffer, fo2_offset = fO2_offset, find_liquidus = find_liquidus)
        # q.put([Results, index])

        # import julia
        from juliacall import Main as jl, convert as jlconvert

        jl.seval("using MAGEMinCalc")

        if Frac_solid is None:
            Frac_solid = False

        comp_julia = jl.seval("Dict")(comp)
        # T_path_C = T_path_C.tolist()
        if T_path_C is not None: 
            T_path_C_julia = jlconvert(jl.Vector[jl.Float64], T_path_C)
        else:
            T_path_C_julia = T_path_C

        if P_path_bar is not None:
            P_path_bar_julia = jlconvert(jl.Vector[jl.Float64], P_path_bar)
        else:
            P_path_bar_julia = P_path_bar

        Results = jl.MAGEMinCalc.path(
            comp=comp_julia, T_start_C=T_start_C, T_end_C=T_end_C, dt_C=dt_C,
            T_C=T_C, P_start_bar=P_start_bar, P_end_bar=P_end_bar, dp_bar=dp_bar,
            P_bar=P_bar, T_path_C=T_path_C_julia, P_path_bar=P_path_bar_julia, frac_xtal=Frac_solid,
            Model=Model, fo2_buffer=fO2_buffer, fo2_offset=fO2_offset, find_liquidus=find_liquidus
        )
        # Results = jl.pyconvert(dict, Results)
        Results_df = dict(Results)
        q.put([Results_df, index])
        # except:
        #     q.put([])
        # return




import numpy as np
import pandas as pd
from PetThermoTools.GenFuncs import *
from PetThermoTools.Plotting import *
from PetThermoTools.MELTS import *
# try:
#     from PetThermoTools.Holland import *
# except:
#     pass
import multiprocessing
from multiprocessing import Queue
from multiprocessing import Process
from tqdm.notebook import tqdm, trange

def equilibrate_multi(cores = None, Model = None, bulk = None, T_C = None, P_bar = None, 
                      Fe3Fet_Liq = None, H2O_Liq = None, CO2_Liq = None, fO2_buffer = None, fO2_offset = None,
                      timeout = None, copy_columns = None, Suppress = None):

    T_C        = to_float(T_C)

    P_bar = to_float(P_bar)

    H2O_Liq = to_float(H2O_Liq)
    CO2_Liq = to_float(CO2_Liq)
    Fe3Fet_Liq = to_float(Fe3Fet_Liq)
    fO2_offset = to_float(fO2_offset)

    if fO2_buffer is not None:
        if fO2_buffer != "NNO":
            if fO2_buffer != "FMQ":
                raise Warning("fO2 buffer specified is not an allowed input. This argument can only be 'FMQ' or 'NNO' \n if you want to offset from these buffers use the 'fO2_offset' argument.")

    if "MELTS" not in Model:
        if fO2_buffer == "FMQ":
            fO2_buffer = "qfm"
        if fO2_buffer == "NNO":
            fO2_buffer = "nno"

    if "MELTS" in Model:
        try:
            from meltsdynamic import MELTSdynamic
        except:
            Warning('alphaMELTS for Python files are not on the python path. \n Please add these files to the path running \n import sys \n sys.path.append(r"insert_your_path_to_melts_here") \n You are looking for the location of the meltsdynamic.py file')

    comp = bulk.copy()

    if Model is None:
        Model = "MELTSv1.0.2"

    # if Model == "Holland":
    #     import pyMAGEMINcalc as MM
    #     print('pyMAGEMinCalc version: ' + str(MM.__version__))

    if cores is None:
        cores = multiprocessing.cpu_count()

    # if comp is entered as a pandas series, it must first be converted to a dict
    if type(comp) == pd.core.series.Series:
        comp = comp.to_dict()

    # ensure the bulk composition has the correct headers etc.
    comp = comp_fix(Model = Model, comp = comp, Fe3Fet_Liq = Fe3Fet_Liq, H2O_Liq = H2O_Liq, CO2_Liq=CO2_Liq)

    if type(comp) == dict:
        if comp['H2O_Liq'] == 0.0 and "MELTS" in Model:
            raise Warning("Adding small amounts of H$_{2}$O may improve the ability of MELTS to accurately reproduce the saturation of oxide minerals. Additionally, sufficient H$_{2}$O is required in the model for MELTS to predict the crystallisation of apatite, rather than whitlockite.")

        if comp['Fe3Fet_Liq'] == 0.0 and "MELTS" in Model and fO2_buffer is None:
            raise Warning("MELTS often fails to produce any results when no ferric Fe is included in the starting composition and an fO2 buffer is not set.")


    if type(P_bar) != np.ndarray:
        if type(comp) == pd.core.frame.DataFrame:
            P_bar = np.zeros(len(comp['SiO2_Liq'])) + P_bar
        else:
            P_bar = np.array([P_bar])

    if type(T_C) != np.ndarray:
        if type(comp) == pd.core.frame.DataFrame:
            T_C = np.zeros(len(comp['SiO2_Liq'])) + T_C
        else:
            T_C = np.array([T_C])

    if "MELTS" in Model:
        if type(comp) == pd.core.frame.DataFrame:
            A = len(P_bar)//cores
            B = len(P_bar) % cores

            if A > 0:
                Group = np.zeros(A) + cores
                if B > 0:
                    Group = np.append(Group, B)
            else:
                Group = np.array([B])

            qs = []
            q = Queue()

            for j in tqdm(range(len(Group))):
                ps = []
                for i in range(int(cores*j), int(cores*j + Group[j])):
                    p = Process(target = equilibrate, args = (q, i), kwargs = {'Model': Model, 'P_bar': P_bar[i], 'T_C': T_C[i], 'comp': comp.loc[i].to_dict(), 'fO2_buffer': fO2_buffer, 'fO2_offset': fO2_offset, 'Suppress': Suppress})

                    ps.append(p)
                    p.start()

                if timeout is None:
                    TIMEOUT = 20
                else:
                    TIMEOUT = timeout

                start = time.time()
                first = True
                for p in ps:
                    if time.time() - start < TIMEOUT - 10:
                        try:
                            ret = q.get(timeout = TIMEOUT - (time.time()-start) + 10)
                        except:
                            ret = []
                    else:
                        if first is True:
                            print('Timeout Reached - this likely indicates the calculation failed. \n You can try increasing the timeout limit using the "timeout" kwarg.')
                            first = False
                        try:
                            ret = q.get(timeout = 10)
                        except:
                            ret = []

                    qs.append(ret)

                TIMEOUT = 5
                start = time.time()
                for p in ps:
                    if p.is_alive():
                        while time.time() - start <= TIMEOUT:
                            if not p.is_alive():
                                p.join()
                                p.terminate()
                                break
                            time.sleep(.1)
                        else:
                            p.terminate()
                            p.join(5)
                    else:
                        p.join()
                        p.terminate()

            Output = {}
            Affinity = {}
            if len(qs) > 1:
                for i in range(len(qs)):
                    if len(qs[i])>0:
                        Results, index = qs[i]

                        try:
                            Output[str(index)] = Results[0]
                            Affinity[str(index)] = Results[1]
                        except:
                            continue

                if "MELTS" in Model:
                    Output = stich(Output, multi = True, Model = "MELTS")

                int_keys = list(map(int, Output.keys()))

                # Determine the size of the resulting DataFrame
                # max_index = max(int_keys)
                max_index = len(comp['SiO2_Liq'])

                # Initialize an empty list to maintain column order
                all_columns = []
                seen_columns = set()

                for df in Output.values():
                    for col in df['All'].columns:
                        if col not in seen_columns:
                            all_columns.append(col)
                            seen_columns.add(col)

                Combined = pd.DataFrame(np.nan, index=range(max_index), columns=all_columns)

                # Populate the result DataFrame
                for key, df in Output.items():
                    try:
                        Combined.loc[int(key)] = df['All'].iloc[0]
                    except:
                        continue

                # Affinity combined
                int_keys = list(map(int, Affinity.keys()))

                # Determine the size of the resulting DataFrame
                # max_index = max(int_keys)
                max_index = len(comp['SiO2_Liq'])

                # Initialize an empty list to maintain column order
                all_columns = []
                seen_columns = set()

                for d in Affinity:
                    for col in Affinity[d]:
                        if col not in seen_columns:
                            all_columns.append(col)
                            seen_columns.add(col)
                    
                    Affinity[d] = pd.Series(Affinity[d])

                Af_Combined = pd.DataFrame(np.nan, index=range(max_index), columns=all_columns)

                # Populate the result DataFrame
                for d in Affinity:
                    try:
                        df = Affinity[d].to_frame().transpose()
                        Af_Combined.loc[int(d)] = df.iloc[0]
                    except:
                        continue
                
            else:
                if len(qs[0]) > 0:
                    Results, index = qs[0]
                    Output = Results[0]
                    Affinity = Results[1]
                
                    if "MELTS" in Model:
                        Output = stich(Output, Model = Model)

        if type(comp) != pd.core.frame.DataFrame:
            A = len(P_bar)//cores
            B = len(P_bar) % cores

            if A > 0:
                Group = np.zeros(A) + cores
                if B > 0:
                    Group = np.append(Group, B)
            else:
                Group = np.array([B])

            qs = []
            q = Queue()

            for j in tqdm(range(len(Group))):
                ps = []
                for i in range(int(cores*j), int(cores*j + Group[j])):
                    p = Process(target = equilibrate, args = (q, i), kwargs = {'Model': Model, 'P_bar': P_bar[i], 'T_C': T_C[i], 'comp': comp, 'fO2_buffer': fO2_buffer, 'fO2_offset': fO2_offset, 'Suppress': Suppress})

                    ps.append(p)
                    p.start()

                if timeout is None:
                    TIMEOUT = 20
                else:
                    TIMEOUT = timeout

                start = time.time()
                first = True
                for p in ps:
                    if time.time() - start < TIMEOUT - 10:
                        try:
                            ret = q.get(timeout = TIMEOUT - (time.time()-start) + 10)
                        except:
                            ret = []
                    else:
                        if first is True:
                            print('Timeout Reached - this likely indicates the calculation failed. \n You can try increasing the timeout limit using the "timeout" kwarg.')
                            first = False
                        try:
                            ret = q.get(timeout = 10)
                        except:
                            ret = []

                    qs.append(ret)

                TIMEOUT = 5
                start = time.time()
                for p in ps:
                    if p.is_alive():
                        while time.time() - start <= TIMEOUT:
                            if not p.is_alive():
                                p.join()
                                p.terminate()
                                break
                            time.sleep(.1)
                        else:
                            p.terminate()
                            p.join(5)
                    else:
                        p.join()
                        p.terminate()

            Output = {}
            Affinity = {}
            if len(qs) > 1:
                for i in range(len(qs)):
                    if len(qs[i])>0:
                        Results, index = qs[i]

                        try:
                            Output[str(index)] = Results[0]
                            Affinity[str(index)] = Results[1]
                        except:
                            continue

                if "MELTS" in Model:
                    Output = stich(Output, multi = True, Model = Model)

                # Output combined
                int_keys = list(map(int, Output.keys()))

                # Determine the size of the resulting DataFrame
                # max_index = max(int_keys)
                max_index = len(P_bar)

                # Initialize an empty list to maintain column order
                all_columns = []
                seen_columns = set()

                for df in Output.values():
                    for col in df['All'].columns:
                        if col not in seen_columns:
                            all_columns.append(col)
                            seen_columns.add(col)

                Combined = pd.DataFrame(np.nan, index=range(max_index), columns=all_columns)

                # Populate the result DataFrame
                for key, df in Output.items():
                    try:
                        Combined.loc[int(key)] = df['All'].iloc[0]
                    except:
                        continue

                # Affinity combined
                int_keys = list(map(int, Affinity.keys()))

                # Determine the size of the resulting DataFrame
                # max_index = max(int_keys)
                max_index = len(P_bar)

                # Initialize an empty list to maintain column order
                all_columns = []
                seen_columns = set()

                for d in Affinity:
                    for col in Affinity[d]:
                        if col not in seen_columns:
                            all_columns.append(col)
                            seen_columns.add(col)
                    Affinity[d] = pd.Series(Affinity[d])

                Af_Combined = pd.DataFrame(np.nan, index=range(max_index), columns=all_columns)

                # Populate the result DataFrame
                for d in Affinity:
                    try:
                        df = Affinity[d].to_frame().transpose()
                        Af_Combined.loc[int(d)] = df.iloc[0]
                    except:
                        continue
            else:
                # try:
                Results, index = qs[0]
                Output = Results[0]
                Affinity = pd.DataFrame([Results[1]])
                Af_Combined = Affinity.copy()
            
                if "MELTS" in Model:
                    Output = stich(Output, Model = Model)

                Combined = Output['All'].copy()
                # except:
                #     return Output

        if copy_columns is not None:
            if type(copy_columns) == str:
                Combined.insert(0, copy_columns, comp[copy_columns])
            elif type(copy_columns) == list:
                j = 0
                for i in copy_columns:
                    Combined.insert(j, i, comp[i])
                    j = j + 1

        Af_Combined = Af_Combined.add_suffix('_affinity')
        Combined = pd.concat([Combined, Af_Combined], axis = 1)
        return Combined
    else:
        from juliacall import Main as jl, convert as jlconvert

        jl.seval("using MAGEMinCalc")

        comp['O'] = comp['Fe3Fet_Liq']*(((159.59/2)/71.844)*comp['FeOt_Liq'] - comp['FeOt_Liq'])

        if Model == "Weller2024":
            bulk = comp[['SiO2_Liq', 'Al2O3_Liq', 'CaO_Liq', 'MgO_Liq', 'FeOt_Liq', 'K2O_Liq', 'Na2O_Liq', 'TiO2_Liq', 'O', 'Cr2O3_Liq']].astype(float).values
        else:
            bulk = comp[['SiO2_Liq', 'Al2O3_Liq', 'CaO_Liq', 'MgO_Liq', 'FeOt_Liq', 'K2O_Liq', 'Na2O_Liq', 'TiO2_Liq', 'O', 'Cr2O3_Liq', 'H2O_Liq']].astype(float).values

        print(np.shape(bulk))
        bulk_jl = jl.seval("collect")(bulk)

        if type(T_C) == np.ndarray:
            T_C = jl.seval("collect")(T_C)
        if type(P_bar) == np.ndarray:
            P_kbar = jl.seval("collect")(P_bar/1000.0)
        else:
            P_kbar = P_bar/1000.0
        if type(fO2_offset) == np.ndarray:
            fO2_offset = jl.seval("collect")(fO2_offset)

        Output = jl.MAGEMinCalc.equilibrate(bulk = bulk_jl, P_kbar = P_kbar, T_C = T_C, fo2_buffer = fO2_buffer, fo2_offset = fO2_offset, Model = Model)
        Output = dict(Output)
        Combined = stich(Output, Model = Model)

        if copy_columns is not None:
            if type(copy_columns) == str:
                Combined['All'].insert(0, copy_columns, comp[copy_columns])
                # Af_Combined.insert(0, copy_columns, comp[copy_columns])
            elif type(copy_columns) == list:
                j = 0
                for i in copy_columns:
                    Combined['All'].insert(j, i, comp[i])
                    # Af_Combined.insert(j, i, comp[i])
                    j = j + 1

        return Combined['All']

def findCO2_multi(cores = None, Model = None, bulk = None, T_initial_C = None, P_bar = None, Fe3Fet_Liq = None, H2O_Liq = None, fO2_buffer = None, fO2_offset = None):

    try:
        from meltsdynamic import MELTSdynamic
    except:
        Warning('alphaMELTS for Python files are not on the python path. \n Please add these files to the path running \n import sys \n sys.path.append(r"insert_your_path_to_melts_here") \n You are looking for the location of the meltsdynamic.py file')

    if fO2_buffer is not None:
        if fO2_buffer != "NNO":
            if fO2_buffer != "FMQ":
                raise Warning("fO2 buffer specified is not an allowed input. This argument can only be 'FMQ' or 'NNO' \n if you want to offset from these buffers use the 'fO2_offset' argument.")

    if "MELTS" not in Model:
        if fO2_buffer == "FMQ":
            fO2_buffer = "qfm"
        if fO2_buffer == "NNO":
            fO2_buffer = "nno"

    comp = bulk.copy()

    if Model is None:
        Model = "MELTSv1.0.2"

    if Model == "Holland":
        import pyMAGEMINcalc as MM

    # if comp is entered as a pandas series, it must first be converted to a dict
    if type(comp) == pd.core.series.Series:
        comp = comp.to_dict()

    # ensure the bulk composition has the correct headers etc.
    comp = comp_fix(Model = Model, comp = comp, Fe3Fet_Liq = Fe3Fet_Liq, H2O_Liq = H2O_Liq)

    if type(comp) != dict or type(P_bar) == np.ndarray:
        if type(comp) != dict:
            T_Liq_C = np.zeros(len(comp['SiO2_Liq'].values))
            H2O_melt = np.zeros(len(comp['SiO2_Liq'].values))
            CO2_melt = np.zeros(len(comp['SiO2_Liq'].values))
            index = np.zeros(len(comp['SiO2_Liq'].values)) - 1
        else:
            T_Liq_C = np.zeros(len(P_bar))
            H2O_melt = np.zeros(len(P_bar))
            CO2_melt = np.zeros(len(P_bar))
            index = np.zeros(len(P_bar)) - 1
    else:
        T_Liq_C = 0
        H2O_melt = 0
        CO2_melt = 0
        index = 0

    if T_initial_C is None:
        if type(comp) != dict or type(P_bar) == np.ndarray:
            if type(comp) != dict:
                T_initial_C = np.zeros(len(comp['SiO2_Liq'].values)) + 1300.0
            else:
                T_initial_C = np.zeros(len(P_bar)) + 1300.0
        else:
            T_initial_C = np.array([1300.0])

    if type(T_initial_C) == int or type(T_initial_C) == float:
        T_initial_C = np.array([T_initial_C])

    if cores is None:
        cores = multiprocessing.cpu_count()

    if type(P_bar) == np.ndarray:
        A = len(P_bar)//cores
        B = len(P_bar) % cores
    elif type(P_bar) != np.ndarray and type(comp) == dict:
        A = 0
        B = 1
    else:
        A = len(comp['SiO2_Liq'])//cores
        B = len(comp['SiO2_Liq']) % cores

    if A > 0:
        Group = np.zeros(A) + cores
        if B > 0:
            Group = np.append(Group, B)
    else:
        Group = np.array([B])

    qs = []
    q = Queue()

    for j in tqdm(range(len(Group))):
        ps = []
        for i in range(int(cores*j), int(cores*j + Group[j])):
            if type(comp) == dict:
                if type(P_bar) == np.ndarray:
                    p = Process(target = findCO2, args = (q, i), kwargs = {'Model': Model, 'P_bar': P_bar[i], 'T_initial_C': T_initial_C[i], 'comp': comp, 'fO2_buffer': fO2_buffer, 'fO2_offset': fO2_offset})
                else:
                    p = Process(target = findCO2, args = (q, i), kwargs = {'Model': Model, 'P_bar': P_bar, 'T_initial_C': T_initial_C[i], 'comp': comp, 'fO2_buffer': fO2_buffer, 'fO2_offset': fO2_offset})
            else:
                if type(P_bar) == np.ndarray:
                    p = Process(target = findCO2, args = (q, i), kwargs = {'Model': Model, 'P_bar': P_bar[i], 'T_initial_C': T_initial_C[i], 'comp': comp.loc[i].to_dict(), 'fO2_buffer': fO2_buffer, 'fO2_offset': fO2_offset})
                else:
                    p = Process(target = findCO2, args = (q, i), kwargs = {'Model': Model, 'P_bar': P_bar, 'T_initial_C': T_initial_C[i], 'comp': comp.loc[i].to_dict(), 'fO2_buffer': fO2_buffer, 'fO2_offset': fO2_offset})

            ps.append(p)
            p.start()

        TIMEOUT = 240

        start = time.time()
        for p in ps:
            if time.time() - start < TIMEOUT - 10:
                try:
                    ret = q.get(timeout = TIMEOUT - (time.time()-start) + 10)
                except:
                    ret = []
            else:
                try:
                    ret = q.get(timeout = 10)
                except:
                    ret = []

            qs.append(ret)

        TIMEOUT = 5
        start = time.time()
        for p in ps:
            if p.is_alive():
                while time.time() - start <= TIMEOUT:
                    if not p.is_alive():
                        p.join()
                        p.terminate()
                        break
                    time.sleep(.1)
                else:
                    p.terminate()
                    p.join(5)
            else:
                p.join()
                p.terminate()

    if type(comp) != dict or type(P_bar) == np.ndarray:
        for i in range(len(qs)):
            if len(qs[i])>0:
                T_Liq_C[i], H2O_melt[i], CO2_melt[i], index[i] = qs[i]

        T_Liq = np.zeros(len(T_Liq_C))
        H2O = np.zeros(len(T_Liq_C))
        CO2 = np.zeros(len(T_Liq_C))

        for i in range(len(index)):
            if len(CO2[index == i]) > 0:
                T_Liq[i] = T_Liq_C[index == i]
                H2O[i] = H2O_melt[index == i]
                CO2[i] = CO2_melt[index == i]
    else:
        T_Liq, H2O, CO2, index = qs[0]

    return T_Liq, H2O, CO2

def findLiq_multi(cores = None, Model = None, bulk = None, T_initial_C = None, P_bar = None, 
                  Fe3Fet_Liq = None, H2O_Liq = None, CO2_Liq = None, fO2_buffer = None, fO2_offset = None, Affinity = False):
    '''
    Perform multiple liquidus temperature (findLiq) calculations in parallel.

    Parameters:
    ----------
    cores: int, optional
        Number of processes to run in parallel. If not provided, the number of available CPU cores is used.

    Model: str, optional
        The thermodynamic model to use for calculations: "MELTS..." or "Holland". 
        If "MELTS..." is selected, the version needs to be specified (e.g., "MELTSv1.0.2", "MELTSv1.1.0", "MELTSv1.2.0", or "pMELTS").
        Default is "MELTSv1.0.2".

    bulk: pd.DataFrame
        A dataframe containing the oxide compositions required for the calculations.

    T_initial_C: float or np.ndarray, optional
        Initial guess for the temperature in Celsius for the findLiq calculations. If not provided, the default is 1300°C.

    P_bar: float or np.ndarray
        Pressure (in bars) for the calculations. Can be a single value or an array with length matching the number of compositions in the bulk.

    Fe3Fet_Liq: float or np.ndarray, optional
        Fe³⁺/total Fe ratio in the liquid. If the composition is provided as a dictionary, this can be an array to vary the Fe³⁺/Fe ratio across compositions. 
        If the composition is a DataFrame, a single value or array (matching the number of compositions) can be used. 
        If not provided, the redox state must be defined elsewhere (e.g., via the fO2 buffer or directly in the bulk variable).

    H2O_Liq: float or np.ndarray, optional
        Initial water content in the liquid phase. Can be a single value for all compositions or an array with values for each composition. 
        If not provided, H2O content must be included in the composition.

    CO2_Liq: float or np.ndarray, optional
        Initial CO₂ content in the liquid phase. Like H2O_Liq, this can be a single value or an array matching the number of compositions.

    fO2_buffer: str, optional
        Oxygen fugacity buffer to control oxidation states during crystallization or decompression. Options include "FMQ" (Fayalite-Magnetite-Quartz) or "NNO" (Nickel-Nickel Oxide).

    fO2_offset: float or np.ndarray, optional
        Offset from the specified fO2 buffer (in log units). Can be a single value or an array.

    Affinity: bool, optional
        If True, the function will return additional information related to chemical affinity (interaction between phases) during crystallization. Default is False.

    Returns:
    ----------
    Res: pd.DataFrame
        DataFrame containing liquidus temperature, liquidus phase, fluid saturation, and normalized melt chemistry.

    Af_Combined: pd.DataFrame, optional
        If Affinity is True, an additional DataFrame containing chemical affinity data for each sample is returned.
    '''
    try:
        from meltsdynamic import MELTSdynamic
    except:
        Warning('alphaMELTS for Python files are not on the python path. \n Please add these files to the path running \n import sys \n sys.path.append(r"insert_your_path_to_melts_here") \n You are looking for the location of the meltsdynamic.py file')

    comp = bulk.copy()

    if fO2_buffer is not None:
        if fO2_buffer != "NNO":
            if fO2_buffer != "FMQ":
                raise Warning("fO2 buffer specified is not an allowed input. This argument can only be 'FMQ' or 'NNO' \n if you want to offset from these buffers use the 'fO2_offset' argument.")

    if "MELTS" not in Model:
        if fO2_buffer == "FMQ":
            fO2_buffer = "qfm"
        if fO2_buffer == "NNO":
            fO2_buffer = "nno"

    if Model is None:
        Model = "MELTSv1.0.2"

    # if Model == "Holland":
    #     import pyMAGEMINcalc as MM

    # if comp is entered as a pandas series, it must first be converted to a dict
    if type(comp) == pd.core.series.Series:
        comp = comp.to_dict()

    # ensure the bulk composition has the correct headers etc.
    comp = comp_fix(Model = Model, comp = comp, Fe3Fet_Liq = Fe3Fet_Liq, H2O_Liq = H2O_Liq, CO2_Liq = CO2_Liq)

    if type(comp) != dict or type(P_bar) == np.ndarray:
        if type(comp) != dict:
            T_Liq = np.zeros(len(comp['SiO2_Liq'].values))
            T_in = np.zeros(len(comp['SiO2_Liq'].values))
            index = np.zeros(len(comp['SiO2_Liq'].values)) - 1
        else:
            T_Liq = np.zeros(len(P_bar))
            T_in = np.zeros(len(P_bar))
            index = np.zeros(len(P_bar)) - 1
    else:
        T_Liq = 0
        T_in = 0
        index = 0

    if T_initial_C is None:
        if type(comp) != dict or type(P_bar) == np.ndarray:
            if type(comp) != dict:
                T_initial_C = np.zeros(len(comp['SiO2_Liq'].values)) + 1300.0
            else:
                T_initial_C = np.zeros(len(P_bar)) + 1300.0
        else:
            T_initial_C = np.array([1300.0])
    else:
        if type(T_initial_C) == int or type(T_initial_C) == float:
            if type(comp) != dict or type(P_bar) == np.ndarray:
                if type(comp) != dict:
                    T_initial_C = np.zeros(len(comp['SiO2_Liq'].values)) + T_initial_C
                else:
                    T_initial_C = np.zeros(len(P_bar)) + T_initial_C
            else:
                T_initial_C = np.array([T_initial_C])
            
    if "MELTS" in Model:

        if cores is None:
            cores = multiprocessing.cpu_count()

        if type(P_bar) == np.ndarray:
            A = len(P_bar)//cores
            B = len(P_bar) % cores
        elif type(P_bar) != np.ndarray and type(comp) == dict:
            A = 0
            B = 1
        else:
            A = len(comp['SiO2_Liq'])//cores
            B = len(comp['SiO2_Liq']) % cores

        if A > 0:
            Group = np.zeros(A) + cores
            if B > 0:
                Group = np.append(Group, B)
        else:
            Group = np.array([B])

        qs = []
        q = Queue()
        for j in tqdm(range(len(Group))):
            ps = []
            for i in range(int(cores*j), int(cores*j + Group[j])):
                if type(comp) == dict:
                    if type(P_bar) == np.ndarray:
                        p = Process(target = findLiq, args = (q, i), kwargs = {'Model': Model, 'P_bar': P_bar[i], 'T_initial_C': T_initial_C[i], 'comp': comp, 'fO2_buffer': fO2_buffer, 'fO2_offset': fO2_offset, 'Affinity': Affinity})
                    else:
                        p = Process(target = findLiq, args = (q, i), kwargs = {'Model': Model, 'P_bar': P_bar, 'T_initial_C': T_initial_C[i], 'comp': comp, 'fO2_buffer': fO2_buffer, 'fO2_offset': fO2_offset, 'Affinity': Affinity})
                else:
                    if type(P_bar) == np.ndarray:
                        if len(comp['SiO2_Liq']) != len(P_bar):
                            raise Warning("The length of your composition and pressure variables are different. Please correct this before running the code.")

                        p = Process(target = findLiq, args = (q, i), kwargs = {'Model': Model, 'P_bar': P_bar[i], 'T_initial_C': T_initial_C[i], 'comp': comp.loc[i].to_dict(), 'fO2_buffer': fO2_buffer, 'fO2_offset': fO2_offset, 'Affinity': Affinity})
                    else:
                        p = Process(target = findLiq, args = (q, i), kwargs = {'Model': Model, 'P_bar': P_bar, 'T_initial_C': T_initial_C[i], 'comp': comp.loc[i].to_dict(), 'fO2_buffer': fO2_buffer, 'fO2_offset': fO2_offset, 'Affinity': Affinity})

                ps.append(p)
                p.start()

            TIMEOUT = 60

            start = time.time()
            for p in ps:
                if time.time() - start < TIMEOUT - 10:
                    try:
                        ret = q.get(timeout = TIMEOUT - (time.time()-start) + 10)
                    except:
                        ret = []
                else:
                    try:
                        ret = q.get(timeout = 10)
                    except:
                        ret = []

                qs.append(ret)

            TIMEOUT = 5
            start = time.time()
            for p in ps:
                if p.is_alive():
                    while time.time() - start <= TIMEOUT:
                        if not p.is_alive():
                            p.join()
                            p.terminate()
                            break
                        time.sleep(.1)
                    else:
                        p.terminate()
                        p.join(5)
                else:
                    p.join()
                    p.terminate()

        Affin = {}
        if type(comp) != dict or type(P_bar) == np.ndarray:
            Results = pd.DataFrame(data = np.zeros((len(T_Liq), 17)), columns = ['T_Liq', 'liquidus_phase', 'fluid_saturated', 'SiO2', 'TiO2', 'Al2O3', 'Fe2O3', 'Cr2O3', 'FeO', 'MnO', 'MgO', 'CaO', 'Na2O', 'K2O', 'P2O5', 'H2O', 'CO2'])
            for i in range(len(qs)):
                if len(qs[i])>0:
                    if Affinity is True:
                        Res, Af, index = qs[i]
                        Affin[str(index)] = Af
                    else:
                        Res, index = qs[i]

                    Results.loc[index,:] = Res
        else:
            if Affinity is True:
                Results, Affin, index = qs[0]
            else:
                Results, index = qs[0]

        Res = comp_fix(Model = Model, comp = Results)
        if type(Res) == dict:
            Res = pd.DataFrame.from_dict(Res, orient = "index").T

        Res = Res[['T_Liq', 'liquidus_phase', 'fluid_saturated', 
        'SiO2_Liq', 'TiO2_Liq', 'Al2O3_Liq', 'Cr2O3_Liq', 'FeOt_Liq','MnO_Liq', 'MgO_Liq', 'CaO_Liq',
        'Na2O_Liq', 'K2O_Liq', 'P2O5_Liq', 'H2O_Liq', 'CO2_Liq', 
        'Fe3Fet_Liq', 'Fe2O3', 'FeO', ]]
        
        Res = Res.drop(columns = ['Fe2O3',  'FeO']) # 'Cr2O3',

        if Affinity is True:
            # Affinity combined
            int_keys = list(map(int, Affin.keys()))

            # Determine the size of the resulting DataFrame
            max_index = max(int_keys)

            # Initialize an empty list to maintain column order
            all_columns = []
            seen_columns = set()

            for d in Affin:
                for col in Affin[d]:
                    if col not in seen_columns:
                        all_columns.append(col)
                        seen_columns.add(col)
                
                Affin[d] = pd.Series(Affin[d])

            Af_Combined = pd.DataFrame(np.nan, index=range(max_index + 1), columns=all_columns)

            # Populate the result DataFrame
            for d in Affin:
                try:
                    df = Affin[d].to_frame().transpose()
                    Af_Combined.loc[int(d)] = df.iloc[0]
                except:
                    continue
            
            return Res, Af_Combined
        else:
            return Res
    else:
        # T_Liq = MM.findLiq_multi(P_bar = P_bar, T_initial_C = T_initial_C, comp = comp)
        return "find liquidus calculations are currently not available through the MAGEMin models. This is an issue I'm working to fix as soon as possible."

def findCO2(q, index, *, Model = None, P_bar = None, T_initial_C = None, comp = None, fO2_buffer = None, fO2_offset = None):
    T_Liq = 0
    H2O_Melt = 0
    CO2_Melt = 0

    if "MELTS" in Model:
        #try:
        T_Liq, H2O_Melt, CO2_Melt = findCO2_MELTS(P_bar = P_bar, Model = Model, T_C = T_initial_C, comp = comp, fO2_buffer = fO2_buffer, fO2_offset = fO2_offset)
        q.put([T_Liq, H2O_Melt, CO2_Melt, index])
        return
        #except:
        #    q.put([T_Liq, H2O_Melt, CO2_Melt, index])
        #    return

def findLiq(q, index,*, Model = None, P_bar = None, T_initial_C = None, comp = None, fO2_buffer = None, fO2_offset = None, Affinity = False):
    '''
    Searches for the liquidus of a melt at the pressure specified. Multiple instances of findLiq are typically initiated in parallel.

    Parameters:
    ----------
    q: Multiprocessing Queue instance
        Queue instance to place the output variables in

    index: int
        index of the calculation in the master code (e.g., position within a for loop) to aid indexing results after calculations are complete.

    Model: string
        "MELTS" or "Holland". Dictates whether MELTS or MAGEMin calculations are performed. Default "MELTS".
        Version of melts can be specified by additing "v1.0.2", "v1.1.0", "v1.2.0", or "p" to "MELTS". Default "v.1.0.2".

    P_bar: float
        Specifies the pressure of the calculation (bar).

    T_initial_C: float
        Initial 'guess' temperature for findLiq calculations (degrees C).

    comp: dict
        Dictionary containing all oxide values required for the calculations.

    Returns:
    ----------
    T_Liq: float
        Estimated liquidus temperatures.

    H2O_Melt: float
        Melt water content at the liquidus

    index: int
        index of the calculation

    T_in: float
        Initial temperature of the calculation - used to ensure indexing is working correctly.
    '''

    T_in = T_initial_C

    
    if "MELTS" in Model:
        try:
            if Affinity is True:
                Results, Affin = findLiq_MELTS(P_bar = P_bar, Model = Model, T_C_init = T_initial_C, comp = comp, fO2_buffer = fO2_buffer, fO2_offset = fO2_offset, Affinity = Affinity)
            else:
                Results = findLiq_MELTS(P_bar = P_bar, Model = Model, T_C_init = T_initial_C, comp = comp, fO2_buffer = fO2_buffer, fO2_offset = fO2_offset, Affinity = Affinity)

            if Affinity is True:
                q.put([Results, Affin, index])
            else:
                q.put([Results, index])
            return
        except:
            Results = {} #pd.DataFrame(data = np.zeros((1, 17)), columns = ['T_Liq', 'liquidus_phase', 'fluid_saturated', 'SiO2', 'TiO2', 'Al2O3', 'Fe2O3', 'Cr2O3', 'FeO', 'MnO', 'MgO', 'CaO', 'Na2O', 'K2O', 'P2O5', 'H2O', 'CO2'])
            Affin = {}
            if Affinity is True:
                q.put([Results, Affin, index])
            else:
                q.put([Results, index])
            return

    if Model == "Holland":
        import pyMAGEMINcalc as MM
        #try:
        Results = MM.findLiq(P_bar = P_bar, T_C_init = T_initial_C, comp = comp)#, fO2_buffer = fO2_buffer, fO2_offset = fO2_offset)
        q.put([Results, index, T_in])
        return
        #except:
        #    q.put([T_Liq, H2O_Melt, index, T_in])
        #    return

def equilibrate(q, i,*, Model = None, P_bar = None, T_C = None, comp = None, fO2_buffer = None, fO2_offset = None, Suppress = None):
    Res = equilibrate_MELTS(Model = Model, P_bar = P_bar, T_C = T_C, comp = comp, fO2_buffer = fO2_buffer, fO2_offset = fO2_offset, Suppress = Suppress)
    q.put([Res, i])
    return